import os, sys

from PyQt5.QtWidgets import QLabel, QApplication, QMessageBox, QSizePolicy
from PyQt5.QtGui import QTextCursor, QIntValidator, QDoubleValidator, QPixmap
from PyQt5.QtCore import Qt

import orangecanvas.resources as resources

from orangewidget import gui, widget
from orangewidget.settings import Setting

from oasys2.widget.widget import OWWidget
from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence

from shadow4.physical_models.mlayer.mlayer import MLayer


from oasys2.widget.util.widget_util import EmittingStream
from orangecontrib.shadow4.util.shadow4_objects import MLayerPreProcessorData
from orangecontrib.shadow4.util.shadow4_util import ShadowPhysics

from orangecontrib.shadow4.widgets.gui.plots import plot_data1D, plot_data2D


class OWMLayer(OWWidget):
    name = "MLayer"
    id = "pre_mlayer"
    description = "Calculation of multilayer mirror reflectivity profile"
    icon = "icons/premlayer.png"
    author = "M Sanchez del Rio"
    maintainer_email = "srio@esrf.eu"
    priority = 30
    category = ""
    keywords = ["preprocessor", "pre_mlayer"]

    outputs = [{"name":"MLayerPreProcessorData",
                "type":MLayerPreProcessorData,
                "doc":"MLayer PreProcessor Data",
                "id":"MLayerPreProcessorData"}]

    want_main_area = True

    #
    # Basic
    #
    FILE = Setting("mlayer.dat")
    E_MIN = Setting(5000.0)
    E_MAX = Setting(20000.0)

    flag_graded = Setting(0)

    structure = Setting('[W,B]x50+Si')
    THICKNESS = Setting(33.1)
    GAMMA = Setting(0.483)

    N_PAIRS = Setting(70)

    O_DENSITY = Setting("9.40")
    O_MATERIAL = Setting("Ru")
    ROUGHNESS_ODD = Setting(3.1)

    E_DENSITY = Setting("2.40")
    E_MATERIAL = Setting("B4C")
    ROUGHNESS_EVEN = Setting(3.3)

    S_DENSITY = Setting("2.33")
    S_MATERIAL = Setting("Si")

    #
    # graded
    #
    grade_coeffs_flag = Setting(0)

    AA0 = Setting(1.0)
    AA1 = Setting(0.0)
    AA2 = Setting(0.0)
    AA3 = Setting(0.0)

    ell_p = Setting(10.0)
    ell_q = Setting(3.0)
    ell_theta_deg = Setting(0.8)
    ell_length = Setting(0.1)
    ell_photon_energy = Setting(10000.0)

    #
    # Plots
    #
    plot_flag = Setting(0)
    scan_e_n = Setting(100)
    scan_e_from = Setting(5000.0)
    scan_e_to = Setting(10000.0)
    scan_a0 = Setting(5.0)

    scan_a_n = Setting(100)
    scan_a_from = Setting(0)
    scan_a_to = Setting(10.0)
    scan_e0 = Setting(8000.0)

    graded_depth_text_list = Setting("[\n[10,10,0.5,0,0],\n[15,20,0.6,0,0],\n]")

    IMAGE_WIDTH  = 860
    IMAGE_HEIGHT = 545

    MAX_WIDTH          = 1320
    MAX_HEIGHT         = 700
    CONTROL_AREA_WIDTH = 405
    TABS_AREA_HEIGHT   = 630

    usage_path = os.path.join(resources.package_dirname("orangecontrib.shadow4.widgets.gui"), "misc", "premlayer_usage.png")

    mlayer_instance = None

    def __init__(self):
        super().__init__()

        self.runaction = widget.OWAction("Compute", self)
        self.runaction.triggered.connect(self.compute)
        self.addAction(self.runaction)

        self.setFixedWidth(self.MAX_WIDTH)
        self.setFixedHeight(self.MAX_HEIGHT)

        gui.separator(self.controlArea)

        box0 = gui.widgetBox(self.controlArea, "",orientation="horizontal")

        button = gui.button(box0, self, "Compute", callback=self.compute)
        button.setFixedHeight(35)

        gui.separator(self.controlArea)

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        self.main_tabs = oasysgui.tabWidget(self.mainArea)
        tab_out = oasysgui.createTabPage(self.main_tabs, "Output")

        self.plot_tab = oasysgui.createTabPage(self.main_tabs, "Plots (optional scan)")


        #
        #
        #
        tab_input = oasysgui.createTabPage(tabs_setting, "Basic Settings")
        self.populate_tab_basic_settings(tab_input)

        tab_input_2 = oasysgui.createTabPage(tabs_setting, "Graded-ML")
        self.populate_tab_graded(tab_input_2)

        tab_input_3 = oasysgui.createTabPage(tabs_setting, "Plots")
        self.populate_tab_plots(tab_input_3)

        tab_input_4 = oasysgui.createTabPage(tabs_setting, "Use of the Widget")
        self.populate_tab_use_of_widget(tab_input_4)


        self.shadow_output = oasysgui.textArea()

        out_box = oasysgui.widgetBox(tab_out, "System Output", addSpace=True, orientation="horizontal", height=self.TABS_AREA_HEIGHT)
        out_box.layout().addWidget(self.shadow_output)

        gui.rubber(self.controlArea)

        self.set_visibility()

    def populate_tab_basic_settings(self, tab_input):
        #
        # Basic Settings
        #
        box = gui.widgetBox(tab_input, "", orientation="vertical")

        box_file = oasysgui.widgetBox(box, "", addSpace=True, orientation="horizontal", height=25)
        self.le_FILE = oasysgui.lineEdit(box_file, self, "FILE",
                                         label="File name (for SHADOW): ", addSpace=True, labelWidth=150, orientation="horizontal")
        gui.button(box_file, self, "...", callback=self.selectFile)


        box_e = gui.widgetBox(box, "", orientation="horizontal")
        oasysgui.lineEdit(box_e, self, "E_MIN", label='Min Energy [eV]', tooltip="E_MIN", addSpace=True,
                          valueType=float, labelWidth=125, orientation="horizontal")

        oasysgui.lineEdit(box_e, self, "E_MAX", label='Max', tooltip="E_MAX", addSpace=True,
                          valueType=float, labelWidth=55, orientation="horizontal")



        gui.comboBox(box, self, "flag_graded", label="graded ML", tooltip="flag_graded", addSpace=True,
                     items=['No (Constant)', 'laterally graded', 'depth graded', 'both'],
                     valueType=int, orientation="horizontal", labelWidth=270, callback=self.set_visibility)


        #
        # structure
        #
        self.box_structure = gui.widgetBox(tab_input, "", orientation="vertical")

        oasysgui.lineEdit(self.box_structure, self, "structure", label="ML structure [Odd,Even]xN+Substrate", tooltip="structure", addSpace=True,
                          labelWidth=250, orientation="horizontal", callback=self.set_structure)


        box_thick_gamma = oasysgui.widgetBox(self.box_structure, "", addSpace=True, orientation="horizontal")
        oasysgui.lineEdit(box_thick_gamma, self, "THICKNESS", label="Bilayer thickness [A]", tooltip="THICKNESS",
                          addSpace=True, labelWidth=125, orientation="horizontal", valueType=float)
        oasysgui.lineEdit(box_thick_gamma, self, "GAMMA", label="gamma=even/total", tooltip="GAMMA",
                          addSpace=True, labelWidth=115, orientation="horizontal", valueType=float)


        #
        box = gui.widgetBox(tab_input, "Sublayers", orientation="vertical")

        self.box_number_of_bilayers = gui.widgetBox(box, orientation="horizontal")
        oasysgui.lineEdit(self.box_number_of_bilayers, self, "N_PAIRS", tooltip="N_PAIRS", label="Number of bilayers", addSpace=True,
                    valueType=int, labelWidth=200, orientation="horizontal", callback=self.get_structure)

        # odd sublayer
        box_odd = gui.widgetBox(box, "Odd sublayer", orientation="vertical")
        oasysgui.lineEdit(box_odd, self, "O_MATERIAL", label='Material [formula]', tooltip="O_MATERIAL",
                           addSpace=True, labelWidth=200, orientation="horizontal", callback=self.set_ODensity)
        bb = gui.widgetBox(box_odd, "", orientation="horizontal")
        oasysgui.lineEdit(bb, self, "O_DENSITY", label='Density [g/cm3]', tooltip="O_DENISTY",
                          addSpace=True, valueType=float, labelWidth=105, orientation="horizontal")
        self.box_roughnessO = gui.widgetBox(bb, orientation="horizontal")
        oasysgui.lineEdit(self.box_roughnessO, self, "ROUGHNESS_ODD", label='Roughness [A]', tooltip="ROUGHNESS_ODD",
                          addSpace=True, valueType=float, labelWidth=95, orientation="horizontal")

        # even sublayer
        box_even = gui.widgetBox(box, "Even sublayer", orientation="vertical")
        oasysgui.lineEdit(box_even, self, "E_MATERIAL", label='Material [formula]', tooltip="E_MATERIAL",
                          addSpace=True, labelWidth=200, orientation="horizontal", callback=self.set_EDensity)
        bb = gui.widgetBox(box_even, "", orientation="horizontal")
        oasysgui.lineEdit(bb, self, "E_DENSITY", label='Density [g/cm3]', tooltip="E_DENISTY",
                          addSpace=True, valueType=float, labelWidth=105, orientation="horizontal")
        self.box_roughnessE = gui.widgetBox(bb, orientation="horizontal")
        oasysgui.lineEdit(self.box_roughnessE, self, "ROUGHNESS_EVEN", label='Roughness [A]', tooltip="ROUGHNESS_EVEN",
                          addSpace=True, valueType=float, labelWidth=95, orientation="horizontal")


        # substrate sublayer
        box_substrate = gui.widgetBox(box, "Substrate", orientation="vertical")
        oasysgui.lineEdit(box_substrate, self, "S_MATERIAL", label='Material [formula]', tooltip="S_MATERIAL",
                           addSpace=True, labelWidth=200, orientation="horizontal", callback=self.set_SDensity)
        oasysgui.lineEdit(box_substrate, self, "S_DENSITY", label='Density [g/cm3]', tooltip="S_DENISTY",
                          addSpace=True, valueType=float, labelWidth=200, orientation="horizontal")


    def populate_tab_graded(self, tab_input_2):

        self.box_lateral = gui.widgetBox(tab_input_2, "laterally graded coeffs F=a[0]+a[1] Y+a[2] Y^2+a[3] Y^3",
                                    orientation="vertical")

        gui.comboBox(self.box_lateral, self, "grade_coeffs_flag", tooltip="grade_coeffs_flag",
                     label="Coeffs from", addSpace=True,
                     items=['External definition', 'Calculate for an ellipse'],
                     valueType=int, orientation="horizontal", labelWidth=270, callback=self.set_visibility)

        self.box_lateral_coeffs = gui.widgetBox(self.box_lateral, "laterally graded coeffs F=a[0]+a[1] Y+a[2] Y^2+a[3] Y^3",
                                         orientation="vertical")
        oasysgui.lineEdit(self.box_lateral_coeffs, self, "AA0", tooltip="AA0", label='a[0]', addSpace=True,
                    valueType=float, labelWidth=150, orientation="horizontal")
        oasysgui.lineEdit(self.box_lateral_coeffs, self, "AA1", tooltip="AA1", label='a[1]', addSpace=True,
                     valueType=float, labelWidth=150, orientation="horizontal")
        oasysgui.lineEdit(self.box_lateral_coeffs, self, "AA2", tooltip="AA2", label='a[2]', addSpace=True,
                    valueType=float, labelWidth=150, orientation="horizontal")
        oasysgui.lineEdit(self.box_lateral_coeffs, self, "AA3", tooltip="AA3", label='a[3]', addSpace=True,
                    valueType=float, labelWidth=150, orientation="horizontal")

        self.box_lateral_ellipse = gui.widgetBox(self.box_lateral, "Ellipse parameters", orientation="vertical")
        oasysgui.lineEdit(self.box_lateral_ellipse, self, "ell_p", tooltip="ell_p", label='p [m]', addSpace=True,
                    valueType=float, labelWidth=150, orientation="horizontal")
        oasysgui.lineEdit(self.box_lateral_ellipse, self, "ell_q", tooltip="ell_q", label='q [m]', addSpace=True,
                     valueType=float, labelWidth=150, orientation="horizontal")
        oasysgui.lineEdit(self.box_lateral_ellipse, self, "ell_theta_deg", tooltip="ell_theta_deg", label='grazing angle [deg]',
                          addSpace=True, valueType=float, labelWidth=200, orientation="horizontal")
        oasysgui.lineEdit(self.box_lateral_ellipse, self, "ell_length", tooltip="ell_length", label='lateral length [m]',
                          addSpace=True, valueType=float, labelWidth=200, orientation="horizontal")
        oasysgui.lineEdit(self.box_lateral_ellipse, self, "ell_photon_energy", tooltip="ell_photon_energy", label='main photon energy [eV]',
                          addSpace=True, valueType=float, labelWidth=200, orientation="horizontal")

        #
        # depth graded
        #
        self.depth_graded_box = gui.widgetBox(tab_input_2, "Depth graded: [ [N,thck,gmma,rouE,rouO], ...]", orientation="vertical")
        self.depth_graded_text_area = oasysgui.textArea(readOnly=False)
        self.depth_graded_box.layout().addWidget(self.depth_graded_text_area)
        self.depth_graded_text_area.setText(self.graded_depth_text_list)

    def populate_tab_plots(self, tab_plots):

        box = gui.widgetBox(tab_plots, "Optional scan plots", orientation="vertical")

        gui.comboBox(box, self, "plot_flag", tooltip="plot_flag",
                     label="Scan plot", addSpace=True,
                     items=['No', 'energy-scan', 'grazing angle-scan', 'energy-angle-scan'],
                     valueType=int, orientation="horizontal", labelWidth=270, callback=self.do_plots)

        self.box_plots = gui.widgetBox(tab_plots, "Optional scan plots", orientation="vertical")

        self.box_plot_e = oasysgui.widgetBox(self.box_plots, "", addSpace=True, orientation="horizontal")
        oasysgui.lineEdit(self.box_plot_e, self, "scan_e_from", label="from [eV]", tooltip="scan_e_from",
                          valueType=float, addSpace=True, labelWidth=100, orientation="horizontal")
        oasysgui.lineEdit(self.box_plot_e, self, "scan_e_to", label="to [eV]", tooltip="scan_e_to",
                          valueType=float, addSpace=True, labelWidth=20, orientation="horizontal")
        oasysgui.lineEdit(self.box_plot_e, self, "scan_e_n", label="points", tooltip="scan_e_n",
                          valueType=int, addSpace=True, labelWidth=50, orientation="horizontal")
        self.box_plot_a0 = oasysgui.widgetBox(self.box_plots, "", addSpace=True, orientation="horizontal")
        oasysgui.lineEdit(self.box_plot_a0, self, "scan_a0", label="Fixed angle [deg]", tooltip="scan_a0",
                          valueType=float, addSpace=True, labelWidth=200, orientation="horizontal")

        self.box_plot_a = oasysgui.widgetBox(self.box_plots, "", addSpace=True, orientation="horizontal")
        oasysgui.lineEdit(self.box_plot_a, self, "scan_a_from", label="from [deg]", tooltip="scan_a_from",
                          valueType=float, addSpace=True, labelWidth=100, orientation="horizontal")
        oasysgui.lineEdit(self.box_plot_a, self, "scan_a_to", label="to [deg]", tooltip="scan_a_to",
                          valueType=float, addSpace=True, labelWidth=20, orientation="horizontal")
        oasysgui.lineEdit(self.box_plot_a, self, "scan_a_n", label="points", tooltip="scan_a_n",
                          valueType=int, addSpace=True, labelWidth=50, orientation="horizontal")
        self.box_plot_e0 = oasysgui.widgetBox(self.box_plots, "", addSpace=True, orientation="horizontal")
        oasysgui.lineEdit(self.box_plot_e0, self, "scan_e0", label="Fixed energy [eV]", tooltip="scan_e0",
                          valueType=float, addSpace=True, labelWidth=200, orientation="horizontal")


    def populate_tab_use_of_widget(self, tab_usa):

        tab_usa.setStyleSheet("background-color: white;")
        tab_usa.setStyleSheet("background-color: white;")

        usage_box = oasysgui.widgetBox(tab_usa, "", addSpace=True, orientation="horizontal")

        label = QLabel("")
        label.setAlignment(Qt.AlignCenter)
        label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        label.setPixmap(QPixmap(self.usage_path))

        usage_box.layout().addWidget(label)


    def set_structure(self):
        #
        # parse
        #
        structure = self.structure
        i0 = structure.find('[')
        i1 = structure.find(',')
        i2 = structure.find(']')
        i3 = structure.find('x')
        i4 = structure.find('+')
        if i0 * i2 * i3 * i4 < 0:
            raise Exception("Bad format for multilayer structure %s (must be: %s)" % (structure, ['[W,B]x50+Si']))
        material_O = structure[(i0+1):(i1)]
        material_E = structure[(i1+1):(i2)]
        npair =  int(structure[(i3+1):(i4)])
        material_S = structure[(i4+1)::]

        print("[%s,%s]x%d+%s" % (material_O, material_E, npair, material_S))

        self.N_PAIRS = npair
        self.O_MATERIAL = material_O
        self.E_MATERIAL = material_E
        self.S_MATERIAL = material_S

        self.set_EDensity()
        self.set_ODensity()
        self.set_SDensity()

    def get_structure(self):
        structure = "[%s,%s]x%d+%s" % (self.O_MATERIAL, self.E_MATERIAL, self.N_PAIRS, self.S_MATERIAL)
        self.structure = structure

    def set_visibility(self):
        self.box_lateral_coeffs.setVisible(self.flag_graded in [1,3] and self.grade_coeffs_flag == 0)
        self.box_lateral_ellipse.setVisible(self.flag_graded in [1,3] and self.grade_coeffs_flag == 1)
        self.box_lateral.setVisible(self.flag_graded in [1,3])

        if self.flag_graded in [2,3]:
            self.depth_graded_box.setVisible(True)
            self.box_structure.setVisible(False)
            self.box_number_of_bilayers.setVisible(False)
            self.box_roughnessE.setVisible(False)
            self.box_roughnessO.setVisible(False)
        else:
            self.depth_graded_box.setVisible(False)
            self.box_structure.setVisible(True)
            self.box_number_of_bilayers.setVisible(True)
            self.box_roughnessE.setVisible(True)
            self.box_roughnessO.setVisible(True)


        self.box_plots.setVisible(self.plot_flag > 0)
        if self.plot_flag == 1:
            self.box_plot_e.setVisible(True)
            self.box_plot_a0.setVisible(True)
            self.box_plot_a.setVisible(False)
            self.box_plot_e0.setVisible(False)
        elif self.plot_flag == 2:
            self.box_plot_e.setVisible(False)
            self.box_plot_a0.setVisible(False)
            self.box_plot_a.setVisible(True)
            self.box_plot_e0.setVisible(True)
        elif self.plot_flag == 3:
            self.box_plot_e.setVisible(True)
            self.box_plot_a0.setVisible(False)
            self.box_plot_a.setVisible(True)
            self.box_plot_e0.setVisible(False)

    def set_SDensity(self):
        if not self.S_MATERIAL is None:
            if not self.S_MATERIAL.strip() == "":
                self.S_MATERIAL = self.S_MATERIAL.strip()
                self.S_DENSITY = ShadowPhysics.getMaterialDensity(self.S_MATERIAL)

    def set_EDensity(self):
        if not self.E_MATERIAL is None:
            if not self.E_MATERIAL.strip() == "":
                self.E_MATERIAL = self.E_MATERIAL.strip()
                self.E_DENSITY = ShadowPhysics.getMaterialDensity(self.E_MATERIAL)
                
    def set_ODensity(self):
        if not self.O_MATERIAL is None:
            if not self.O_MATERIAL.strip() == "":
                self.O_MATERIAL = self.O_MATERIAL.strip()
                self.O_DENSITY = ShadowPhysics.getMaterialDensity(self.O_MATERIAL)

    def compute(self):
        try:
            sys.stdout = EmittingStream(textWritten=self.writeStdOut)
            self.shadow_output.setText("")
            self.checkFields()

            if self.flag_graded == 0:
                GRADE_SURFACE = 0
                GRADE_DEPTH = 0
                LIST_N_THICK_GAMMA_ROUGHE_ROUGHO_FROM_TOP_TO_BOTTOM = None
            elif self.flag_graded == 1: # lateral
                if self.grade_coeffs_flag == 0: # external coefficients in m
                    GRADE_SURFACE = 3  # S4 coeffs
                else:
                    GRADE_SURFACE = 4
                GRADE_DEPTH = 0
                LIST_N_THICK_GAMMA_ROUGHE_ROUGHO_FROM_TOP_TO_BOTTOM = None
            elif self.flag_graded == 2:  # depth
                GRADE_SURFACE = 0
                GRADE_DEPTH = 1
                LIST_N_THICK_GAMMA_ROUGHE_ROUGHO_FROM_TOP_TO_BOTTOM = self.depth_graded_text_area.toPlainText()
                self.graded_depth_text_list = self.depth_graded_text_area.toPlainText()
            elif self.flag_graded == 3:  # both
                if self.grade_coeffs_flag == 0: # external coefficients in m
                    GRADE_SURFACE = 3  # S4 coeffs
                else:
                    GRADE_SURFACE = 4
                GRADE_DEPTH = 1
                LIST_N_THICK_GAMMA_ROUGHE_ROUGHO_FROM_TOP_TO_BOTTOM = self.depth_graded_text_area.toPlainText()
                self.graded_depth_text_list = self.depth_graded_text_area.toPlainText()
            else:
                raise NotImplementedError()

            self.mlayer_instance = MLayer.pre_mlayer(
                             FILE=congruence.checkFileName(self.FILE),
                             E_MIN=self.E_MIN,
                             E_MAX=self.E_MAX,
                             S_DENSITY=self.S_DENSITY,
                             S_MATERIAL=self.S_MATERIAL,
                             E_DENSITY=self.E_DENSITY,
                             E_MATERIAL=self.E_MATERIAL,
                             O_DENSITY=self.O_DENSITY,
                             O_MATERIAL=self.O_MATERIAL,
                             N_PAIRS=self.N_PAIRS,
                             THICKNESS=self.THICKNESS,
                             GAMMA=self.GAMMA,
                             ROUGHNESS_EVEN=self.ROUGHNESS_EVEN,
                             ROUGHNESS_ODD=self.ROUGHNESS_ODD,
                             GRADE_SURFACE=GRADE_SURFACE,
                             AA0=self.AA0,
                             AA1=self.AA1,
                             AA2=self.AA2,
                             AA3=self.AA3,
                             ell_p=self.ell_p,
                             ell_q=self.ell_q,
                             ell_theta_grazing_deg=self.ell_theta_deg,
                             ell_length=self.ell_length,
                             ell_photon_energy=self.ell_photon_energy,
                             GRADE_DEPTH=GRADE_DEPTH,
                             LIST_N_THICK_GAMMA_ROUGHE_ROUGHO_FROM_TOP_TO_BOTTOM=LIST_N_THICK_GAMMA_ROUGHE_ROUGHO_FROM_TOP_TO_BOTTOM,
                             )

            # this is for just info
            if self.flag_graded == 1 and self.grade_coeffs_flag == 1:
                _ = self.mlayer_instance._fit_ellipse_laterally_graded_coeffs(verbose=1)

            # plots
            self.do_plots()

            self.send("MLayerPreProcessorData", MLayerPreProcessorData(mlayer_data_file=self.FILE))
        except Exception as exception:
            QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)
            if self.IS_DEVELOP: raise exception

    def checkFields(self):
        congruence.checkDir(self.FILE)
        self.E_MIN  = congruence.checkPositiveNumber(self.E_MIN , "Min Energy")
        self.E_MAX  = congruence.checkStrictlyPositiveNumber(self.E_MAX , "Max Energy")
        congruence.checkLessOrEqualThan(self.E_MIN, self.E_MAX, "Minimum Energy", "Maximum Energy")
        self.S_MATERIAL = ShadowPhysics.checkCompoundName(self.S_MATERIAL)
        self.S_DENSITY = congruence.checkStrictlyPositiveNumber(float(self.S_DENSITY), "Density (substrate)")
        self.E_MATERIAL = ShadowPhysics.checkCompoundName(self.E_MATERIAL)
        self.E_DENSITY = congruence.checkStrictlyPositiveNumber(float(self.E_DENSITY), "Density (even sublayer)")
        self.O_MATERIAL = ShadowPhysics.checkCompoundName(self.O_MATERIAL)
        self.O_DENSITY = congruence.checkStrictlyPositiveNumber(float(self.O_DENSITY), "Density (odd sublayer)")

        if self.flag_graded in [0,1,3]:
            self.N_PAIRS = congruence.checkStrictlyPositiveNumber(int(self.N_PAIRS), "Number of bilayers")
            self.THICKNESS = congruence.checkStrictlyPositiveNumber(float(self.THICKNESS), "bilayer thickness t")
            self.GAMMA = congruence.checkStrictlyPositiveNumber(float(self.GAMMA), "gamma ratio")
            self.ROUGHNESS_EVEN = congruence.checkPositiveNumber(float(self.ROUGHNESS_EVEN), "Roughness even layer")
            self.ROUGHNESS_ODD = congruence.checkPositiveNumber(float(self.ROUGHNESS_ODD), "Roughness odd layer")

        if self.flag_graded in [1,3]:
            self.AA0 = congruence.checkNumber(float(self.AA0), "zero-order coefficient")
            self.AA1 = congruence.checkNumber(float(self.AA1), "linear coefficient")
            self.AA2 = congruence.checkNumber(float(self.AA2), "2nd degree coefficient")
            self.AA3 = congruence.checkNumber(float(self.AA3), "3rd degree coefficient")
            # todo add ellipse

    def selectFile(self):
        self.le_FILE.setText(oasysgui.selectFileFromDialog(self, self.FILE, "Select Output File", file_extension_filter="Data Files (*.dat)"))

    def selectFileDepth(self):
        self.le_FILE_DEPTH.setText(oasysgui.selectFileFromDialog(self, self.FILE_DEPTH, "Open File with list of t_bilayer,gamma,roughness_even,roughness_odd", file_extension_filter="Data Files (*.dat)"))

    def do_plots(self):

        self.set_visibility()

        self.plot_tab.layout().removeItem(self.plot_tab.layout().itemAt(0))

        if self.mlayer_instance is None: return

        if self.plot_flag == 0:
            plot_widget_id = plot_data1D([0], [0], xtitle="", ytitle="")
        elif self.plot_flag == 1:
            energyN = int(self.scan_e_n)
            energy1 = self.scan_e_from
            energy2 = self.scan_e_to
            thetaN = 1
            theta1 = self.scan_a0
            theta2 = self.scan_a0
            R_S_array, R_P_array, energy_array, theta_array = self.mlayer_instance.scan(
                energyN=energyN, energy1=energy1, energy2=energy2,
                thetaN=thetaN, theta1=theta1, theta2=theta2)
            plot_widget_id = plot_data1D(energy_array, R_S_array[:, 0] ** 2, xtitle="Photon energy [eV]",
                                         ytitle="Reflectivity")
        elif self.plot_flag == 2:
            energyN = 1
            energy1 = self.scan_e0
            energy2 = self.scan_e0
            thetaN = int(self.scan_a_n)
            theta1 = self.scan_a_from
            theta2 = self.scan_a_to
            R_S_array, R_P_array, energy_array, theta_array = self.mlayer_instance.scan(
                energyN=energyN, energy1=energy1, energy2=energy2,
                thetaN=thetaN, theta1=theta1, theta2=theta2)
            plot_widget_id = plot_data1D(theta_array, R_S_array[0, :]**2, xtitle="grazing angle [deg]", ytitle="Reflectivity")
        elif self.plot_flag == 3:
            energyN = self.scan_e_n
            energy1 = self.scan_e_from
            energy2 = self.scan_e_to
            thetaN = self.scan_a_n
            theta1 = self.scan_a_from
            theta2 = self.scan_a_to
            R_S_array, R_P_array, energy_array, theta_array = self.mlayer_instance.scan(
                energyN=energyN, energy1=energy1, energy2=energy2,
                thetaN=thetaN, theta1=theta1, theta2=theta2)
            plot_widget_id = plot_data2D(R_S_array**2, energy_array, theta_array, title="title",
                                         xtitle="photon energy [eV]", ytitle="grazing angle [deg]")

        self.plot_tab.layout().addWidget(plot_widget_id)

    def writeStdOut(self, text):
        cursor = self.shadow_output.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        self.shadow_output.setTextCursor(cursor)
        self.shadow_output.ensureCursorVisible()

if __name__ == "__main__":
    app = QApplication(sys.argv)
    w = OWMLayer()
    w.show()
    app.exec_()
    w.saveSettings()



import sys, time
import numpy

from PyQt5 import QtGui
from PyQt5.QtWidgets import QApplication, QMessageBox

from orangewidget import gui
from orangewidget.settings import Setting
from oasys2.widget import gui as oasysgui
from oasys2.widget.util.widget_util import EmittingStream

from orangecontrib.shadow4.widgets.gui.ow_automatic_element import AutomaticElement
from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence
from orangecontrib.shadow4.widgets.gui.plots import plot_multi_data1D
from orangecontrib.shadow4.util.python_script import PythonScript

from shadow4.tools.beamline_tools import focnew, focnew_scan, focnew_scan_full_beamline
from shadow4.tools.logger import set_verbose

class FocNew(AutomaticElement):

    name = "FocNew"
    description = "Tools: FocNew"
    icon = "icons/focnew.png"
    maintainer = "M. Sanchez del Rio and L. Rebuffi"
    maintainer_email = "srio(@at@)esrf.eu"
    priority = 5
    category = "Data Display Tools"
    keywords = ["data", "file", "load", "read"]

    inputs = [("Shadow Data", ShadowData, "set_shadow_data")]

    IMAGE_WIDTH = 860
    IMAGE_HEIGHT = 675

    want_main_area=1
    want_control_area = 1

    mode = Setting(0)
    center_x = Setting(0.0)
    center_z = Setting(0.0)

    y_range_min=Setting(-10.0)
    y_range_max=Setting(10.0)
    y_npoints=Setting(1001)

    plot_beamline = Setting(1)
    npoints_beamline=Setting(11)

    plot_canvas_x = None
    plot_canvas_bl = None

    input_data = None

    def __init__(self):
        super().__init__(show_automatic_box=True)

        gui.button(self.controlArea, self, "Calculate", callback=self.calculate, height=45)

        general_box = oasysgui.widgetBox(self.controlArea, "General Settings", addSpace=True, orientation="vertical")


        general_box1 = oasysgui.widgetBox(general_box, "Center", addSpace=True, orientation="vertical")
        gui.comboBox(general_box1, self, "mode", label="Mode", labelWidth=250,
                                     items=["Center at Origin",
                                            "Center at Barycenter",
                                            "External"],
                                     callback=self.set_visibility, sendSelectedValue=False, orientation="horizontal")
        self.center_box = oasysgui.widgetBox(general_box1, "", addSpace=False, orientation="vertical", height=50)
        self.le_center_x = oasysgui.lineEdit(self.center_box, self, "center_x", "Center X [m]", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_center_z = oasysgui.lineEdit(self.center_box, self, "center_z", "Center Z [m]", labelWidth=260, valueType=float, orientation="horizontal")


        general_box1 = oasysgui.widgetBox(general_box, "Scan range", addSpace=True, orientation="vertical")
        self.yrange_box = oasysgui.widgetBox(general_box1, "", addSpace=False, orientation="vertical", height=100)
        self.le_y_range_min = oasysgui.lineEdit(self.yrange_box, self, "y_range_min", "Y min [m]", labelWidth=250, valueType=float, orientation="horizontal")
        self.le_y_range_max = oasysgui.lineEdit(self.yrange_box, self, "y_range_max", "Y max [m]", labelWidth=250, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.yrange_box, self, "y_npoints", "Points (for focnew scan)", labelWidth=300, valueType=int, orientation="horizontal")


        general_box1 = oasysgui.widgetBox(general_box, "Beamline", addSpace=True, orientation="vertical")
        gui.comboBox(general_box1, self, "plot_beamline", label="Full beamline plot", labelWidth=250,
                                     items=["No", "Yes"],
                                     callback=self.set_visibility, sendSelectedValue=False, orientation="horizontal")
        self.flag_beamline_box = oasysgui.widgetBox(general_box1, "", addSpace=False, orientation="vertical", height=100)
        oasysgui.lineEdit(self.flag_beamline_box, self, "npoints_beamline", "Points inter o.e. (for focnew full beamline)", labelWidth=300, valueType=int, orientation="horizontal")


        gui.separator(self.controlArea, height=200)

        tabs_setting = oasysgui.tabWidget(self.mainArea)
        tabs_setting.setFixedHeight(self.IMAGE_HEIGHT+5)
        tabs_setting.setFixedWidth(self.IMAGE_WIDTH)

        tab_info = oasysgui.createTabPage(tabs_setting, "Focnew Info")
        tab_scan = oasysgui.createTabPage(tabs_setting, "Focnew Scan")
        tab_beamline = oasysgui.createTabPage(tabs_setting, "Focnew Full Beamline")
        tab_out = oasysgui.createTabPage(tabs_setting, "System Output")
        tab_script = oasysgui.createTabPage(tabs_setting, "Script")

        self.focnewInfo = oasysgui.textArea(height=self.IMAGE_HEIGHT-35)
        info_box = oasysgui.widgetBox(tab_info, "", addSpace=True, orientation="horizontal", width = self.IMAGE_WIDTH-20, ) # height = self.IMAGE_HEIGHT-20)
        info_box.layout().addWidget(self.focnewInfo)

        self.image_box = gui.widgetBox(tab_scan, "Scan", addSpace=True, orientation="vertical")
        self.image_box.setFixedHeight(self.IMAGE_HEIGHT-30)
        self.image_box.setFixedWidth(self.IMAGE_WIDTH-20)

        self.beamline_box = gui.widgetBox(tab_beamline, "Beamline", addSpace=True, orientation="vertical")
        self.beamline_box.setFixedHeight(self.IMAGE_HEIGHT-30)
        self.beamline_box.setFixedWidth(self.IMAGE_WIDTH-20)

        self.shadow4_script = PythonScript()
        self.shadow4_script.code_area.setFixedHeight(400)
        script_box = gui.widgetBox(tab_script, "Python script", addSpace=True, orientation="horizontal")
        script_box.layout().addWidget(self.shadow4_script)

        self.shadow_output = oasysgui.textArea(height=self.IMAGE_HEIGHT-35)
        out_box = oasysgui.widgetBox(tab_out, "System Output", addSpace=True, orientation="horizontal", width = self.IMAGE_WIDTH-20, ) # height = self.IMAGE_HEIGHT-20)
        out_box.layout().addWidget(self.shadow_output)

        self.set_visibility()


    def set_visibility(self):
        self.center_box.setVisible(False)
        self.center_box.setVisible(self.mode == 2)

        self.flag_beamline_box.setVisible(False)
        self.flag_beamline_box.setVisible(self.plot_beamline == 1)

    def calculate(self):
        if self.input_data is None:
            self.prompt_exception(ValueError("No input beam"))
            return

        try:
            set_verbose(0)
            sys.stdout = EmittingStream(textWritten=self.write_stdout)

            self.shadow_output.setText("")
            self.focnewInfo.setText("")

            if self.plot_canvas_x is not None:
                self.image_box.layout().removeItem(self.image_box.layout().itemAt(0))
                self.plot_canvas_x.hide()
                self.plot_canvas_x = None

            if self.plot_canvas_bl is not None:
                self.beamline_box.layout().removeItem(self.beamline_box.layout().itemAt(0))
                self.plot_canvas_bl.hide()
                self.plot_canvas_bl = None


            self.set_script()

            self.do_plot_focnew()
            if self.plot_beamline: self.do_plot_beamline()



        except Exception as exception:
            self.prompt_exception(exception)


    def do_plot_focnew(self):

        ticket = focnew(beam=self.input_data.beam, nolost=1, mode=self.mode, center=[self.center_x, self.center_z])

        # info...
        self.focnewInfo.setText(ticket['text'])
        print("list of 6 coeffs: <d**2>, <x d>, <x**2>, <x>**2, <x><d>, <d>**2: ")
        print("AX coeffs: ", ticket["AX"])
        print("AZ coeffs: ", ticket["AZ"])
        print("AT coeffs: ", ticket["AT"])

        # scan...
        y = numpy.linspace(self.y_range_min, self.y_range_max, int(self.y_npoints))

        ylist = [focnew_scan(ticket["AX"], y) * 1e6,
                 focnew_scan(ticket["AZ"], y) * 1e6,
                 focnew_scan(ticket["AT"], y) * 1e6]

        self.plot_canvas_x = plot_multi_data1D(y, ylist,
                                               title="title",
                                               xtitle="Y [m]",
                                               ytitle="<X> or <Z> or <X,Z> [$\mu$m]",
                                               ytitles=["X","Z","X,Z combined"],
                                               flag_common_abscissas=1)

        self.image_box.layout().addWidget(self.plot_canvas_x)

    def do_plot_beamline(self):
        ticket = focnew_scan_full_beamline(self.input_data.beamline, npoints=self.npoints_beamline)

        self.plot_canvas_bl = plot_multi_data1D(ticket['list_y'] + ticket['list_y'],
                                                ticket['list_x'] + ticket['list_z'],
                                                title="title",
                                                xtitle="Y [m]",
                                                ytitle="<H> or <V> [$\mu$m]",
                                                ytitles=ticket['list_x_label'] + ticket['list_z_label'],
                                                flag_common_abscissas=0)


        self.beamline_box.layout().addWidget(self.plot_canvas_bl)


    def set_shadow_data(self, input_data):
        if ShadowCongruence.check_empty_data(input_data):
            self.input_data = input_data.duplicate()
            if self.is_automatic_run: self.calculate()

    def write_stdout(self, text):
        cursor = self.shadow_output.textCursor()
        cursor.movePosition(QtGui.QTextCursor.End)
        cursor.insertText(text)
        self.shadow_output.setTextCursor(cursor)
        self.shadow_output.ensureCursorVisible()

    def set_script(self):

        # script
        try:
            beamline = self.input_data.beamline.duplicate()
            script = beamline.to_python_code()

            indented_script = '\n'.join('    ' + line for line in script.splitlines())

            final_script = "def run_beamline():\n"
            final_script += indented_script
            final_script += "\n    return beam, footprint, beamline"
            final_script += "\n\n"

            dict = {
                    "mode": self.mode,
                    "center_x": self.center_x,
                    "center_z": self.center_z,
                    "y_range_min": self.y_range_min,
                    "y_range_max": self.y_range_max,
                    "y_npoints": int(self.y_npoints),
                    "npoints_beamline": int(self.npoints_beamline),
                    "plot_beamline": self.plot_beamline,
                    }

            script_template = """#
# main
#
import numpy
from shadow4.tools.beamline_tools import focnew, focnew_scan, focnew_scan_full_beamline
from srxraylib.plot.gol import plot, plot_image, plot_image_with_histograms, plot_show

beam, footprint, beamline = run_beamline()

ticket1 = focnew(beam=beam, nolost=1, mode={mode}, center=[{center_x}, {center_z}])
print(ticket1['text'])

y = numpy.linspace({y_range_min}, {y_range_max}, {y_npoints})
x = focnew_scan(ticket1['AX'], y) * 1e6
z = focnew_scan(ticket1['AZ'], y) * 1e6
t = focnew_scan(ticket1['AT'], y) * 1e6
plot(y, x, y, z, y, t, xtitle='Y [m]', ytitle='<x> or <z> [um]', title='at specific optical element', legend=['x','z','combined'])

if {plot_beamline}:
    ticket2 = focnew_scan_full_beamline(beamline, npoints={npoints_beamline})
    plot(ticket2['y'], ticket2['x'], ticket2['y'], ticket2['z'], xtitle='Y [m]', ytitle='<H> or <V> [um]', title='Full beamline', legend=['H','V'])

"""

            final_script += script_template.format_map(dict)

            self.shadow4_script.set_code(final_script)
        except:
            final_script += "\n\n\n# cannot retrieve beamline data from shadow_data"

        self.shadow4_script.set_code(final_script)

if __name__ == "__main__":
    def get_beamline():
        from shadow4.beamline.s4_beamline import S4Beamline

        beamline = S4Beamline()

        # electron beam
        from shadow4.sources.s4_electron_beam import S4ElectronBeam
        electron_beam = S4ElectronBeam(energy_in_GeV=6, energy_spread=0.001, current=0.2)
        electron_beam.set_sigmas_all(sigma_x=3.01836e-05, sigma_y=4.36821e-06, sigma_xp=3.63641e-06,
                                     sigma_yp=1.37498e-06)

        # magnetic structure
        from shadow4.sources.undulator.s4_undulator_gaussian import S4UndulatorGaussian
        source = S4UndulatorGaussian(
            period_length=0.042,  # syned Undulator parameter (length in m)
            number_of_periods=38.571,  # syned Undulator parameter
            photon_energy=5000.0,  # Photon energy (in eV)
            delta_e=4.0,  # Photon energy width (in eV)
            ng_e=100,  # Photon energy scan number of points
            flag_emittance=1,  # when sampling rays: Use emittance (0=No, 1=Yes)
            flag_energy_spread=0,  # when sampling rays: Use e- energy spread (0=No, 1=Yes)
            harmonic_number=1,  # harmonic number
            flag_autoset_flux_central_cone=1,  # value to set the flux peak
            flux_central_cone=681709040139326.4,  # value to set the flux peak
        )

        # light source
        from shadow4.sources.undulator.s4_undulator_gaussian_light_source import S4UndulatorGaussianLightSource
        light_source = S4UndulatorGaussianLightSource(name='GaussianUndulator', electron_beam=electron_beam,
                                                      magnetic_structure=source, nrays=15000, seed=5676561)
        beam = light_source.get_beam()

        beamline.set_light_source(light_source)

        # optical element number XX
        from syned.beamline.shape import Rectangle
        boundary_shape = Rectangle(x_left=-0.001, x_right=0.001, y_bottom=-0.001, y_top=0.001)

        from shadow4.beamline.optical_elements.absorbers.s4_screen import S4Screen
        optical_element = S4Screen(name='Generic Beam Screen/Slit/Stopper/Attenuator', boundary_shape=boundary_shape,
                                   i_abs=0,  # 0=No, 1=prerefl file_abs, 2=xraylib, 3=dabax
                                   i_stop=0, thick=0, file_abs='<specify file name>', material='Au', density=19.3)

        from syned.beamline.element_coordinates import ElementCoordinates
        coordinates = ElementCoordinates(p=27.2, q=0, angle_radial=0, angle_azimuthal=0, angle_radial_out=3.141592654)
        from shadow4.beamline.optical_elements.absorbers.s4_screen import S4ScreenElement
        beamline_element = S4ScreenElement(optical_element=optical_element, coordinates=coordinates, input_beam=beam)

        beam, footprint = beamline_element.trace_beam()

        beamline.append_beamline_element(beamline_element)

        # optical element number XX
        boundary_shape = None

        from shadow4.beamline.optical_elements.mirrors.s4_plane_mirror import S4PlaneMirror
        optical_element = S4PlaneMirror(name='Plane Mirror', boundary_shape=boundary_shape,
                                        f_reflec=1, f_refl=5, file_refl='<none>', refraction_index=0.99999 + 0.001j,
                                        coating_material='Ni', coating_density=8.902, coating_roughness=0)

        from syned.beamline.element_coordinates import ElementCoordinates
        coordinates = ElementCoordinates(p=2.7, q=0, angle_radial=1.563796327, angle_azimuthal=1.570796327,
                                         angle_radial_out=1.563796327)
        movements = None
        from shadow4.beamline.optical_elements.mirrors.s4_plane_mirror import S4PlaneMirrorElement
        beamline_element = S4PlaneMirrorElement(optical_element=optical_element, coordinates=coordinates,
                                                movements=movements, input_beam=beam)

        beam, mirr = beamline_element.trace_beam()

        beamline.append_beamline_element(beamline_element)

        # test plot
        if 0:
            from srxraylib.plot.gol import plot_scatter
            plot_scatter(beam.get_photon_energy_eV(nolost=1), beam.get_column(23, nolost=1),
                         title='(Intensity,Photon Energy)', plot_histograms=0)
            plot_scatter(1e6 * beam.get_column(1, nolost=1), 1e6 * beam.get_column(3, nolost=1),
                         title='(X,Z) in microns')
        return beam, footprint, beamline
    ###############################
    beam, footprint, beamline = get_beamline()

    if 1:
        app = QApplication(sys.argv)
        w = FocNew()
        w.set_shadow_data(ShadowData(beam=beam, footprint=footprint, number_of_rays=0, beamline=beamline))
        w.show()
        app.exec()
        # w.saveSettings()

    if 0:
        #############################################################
        import numpy
        from shadow4.tools.beamline_tools import focnew, focnew_scan, focnew_scan_full_beamline

        # beam = in_object_1.beam  # define your beam
        # beamline = in_object_1.beamline  # define your beamline
        ticket1 = focnew(beam=beam, nolost=1, mode=0, center=[0.000000, 0.000000])
        y = numpy.linspace(-10.000000, 10.000000, 1001)
        x = focnew_scan(ticket1['AX'], y) * 1e6
        z = focnew_scan(ticket1['AZ'], y) * 1e6
        t = focnew_scan(ticket1['AT'], y) * 1e6
        from srxraylib.plot.gol import plot

        plot(y, x, y, z, xtitle='Y [m]', ytitle='<x> or <z> [um]', legend=['x', 'z'])

        ticket2 = focnew_scan_full_beamline(beamline)
        plot(ticket2['y'], ticket2['x'], ticket2['y'], ticket2['z'], xtitle='Y [m]', ytitle='<H> or <V> [um]',
             legend=['H', 'V'])
        #############################################################


import os, sys, numpy

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QLabel, QApplication, QMessageBox, QSizePolicy
from PyQt5.QtGui import QTextCursor, QPixmap

import orangecanvas.resources as resources

from orangewidget import gui, widget
from orangewidget.settings import Setting

from oasys2.widget.widget import OWWidget
from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence
from oasys2.widget.util.widget_util import EmittingStream

from orangecontrib.shadow4.util.shadow4_objects import VlsPgmPreProcessorData
from orangecontrib.shadow4.util.shadow4_util import ShadowPhysics # todo change

class OWVlsPgmCoefficientsCalculator(OWWidget):
    name = "VLS PGM Coefficients Calculator"
    id = "VlsPgmCoefficientsCalculator"
    description = "Calculation of coefficients for VLS PGM"
    icon = "icons/vls_pgm.png"
    author = "Luca Rebuffi"
    maintainer_email = "lrebuffi@anl.gov"
    priority = 100
    category = ""
    keywords = ["oasys", "vls", "pgm"]

    outputs = [{"name":"VLS-PGM PreProcessor Data",
                "type":VlsPgmPreProcessorData,
                "doc":"VLS-PGM PreProcessor Data",
                "id":"VLS-PGM PreProcessor Data"}]

    want_main_area = True

    last_element_distance = Setting(0.0)

    r_a = Setting(0.0)
    r_b = Setting(0.0)
    k = Setting(1000)

    h = Setting(20)

    units_in_use = Setting(0)
    photon_wavelength = Setting(25.0)
    photon_energy = Setting(500.0)

    c = Setting(1.2)
    grating_diffraction_order = Setting(-1)

    new_units_in_use = Setting(0)
    new_photon_wavelength = Setting(25.0)
    new_photon_energy = Setting(500.0)
    new_c_value = Setting(1.2222)
    new_c_flag = Setting(0)

    image_path = os.path.join(resources.package_dirname("orangecontrib.shadow4.widgets.gui"), "misc", "vls_pgm_layout.png")
    usage_path = os.path.join(resources.package_dirname("orangecontrib.shadow4.widgets.gui"), "misc", "vls_pgm_usage.png")

    design_alpha = 0.0
    design_beta = 0.0

    b2 = 0.0
    b3 = 0.0
    b4 = 0.0

    shadow_coeff_0 = 0.0
    shadow_coeff_1 = 0.0
    shadow_coeff_2 = 0.0
    shadow_coeff_3 = 0.0

    d_source_to_mirror = 0.0
    d_source_plane_to_mirror = 0.0
    d_mirror_to_grating = 0.0

    raytracing_alpha = 0.0
    raytracing_beta = 0.0


    def __init__(self):

        super().__init__()

        self.runaction = widget.OWAction("Compute", self)
        self.runaction.triggered.connect(self.compute)
        self.addAction(self.runaction)

        self.setFixedWidth(1170)
        self.setFixedHeight(500)

        gui.separator(self.controlArea)

        box0 = oasysgui.widgetBox(self.controlArea, "", orientation="horizontal")
        #widget buttons: compute, set defaults, help
        button = gui.button(box0, self, "Compute", callback=self.compute)
        button.setFixedHeight(45)
        button = gui.button(box0, self, "Defaults", callback=self.defaults)
        button.setFixedHeight(45)

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(425)

        tab_step_1 = oasysgui.createTabPage(tabs_setting, "Line Density Calculation")
        tab_step_2 = oasysgui.createTabPage(tabs_setting, "Angles Calculation")
        tab_usa = oasysgui.createTabPage(tabs_setting, "Use of the Widget")
        tab_usa.setStyleSheet("background-color: white;")

        usage_box = oasysgui.widgetBox(tab_usa, "", addSpace=True, orientation="horizontal")

        label = QLabel("")
        label.setAlignment(Qt.AlignCenter)
        label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        label.setPixmap(QPixmap(self.usage_path))

        usage_box.layout().addWidget(label)

        box = oasysgui.widgetBox(tab_step_1, "VLS-PGM Layout Parameters", orientation="vertical")

        oasysgui.lineEdit(box, self, "r_a", "Distance Source-Grating [m]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box, self, "r_b", "Distance Grating-Exit Slits [m]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box, self, "h", "Vertical Distance Mirror-Grating [m]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box, self, "last_element_distance", "Distance Source-Last Image Plane\nbefore Mirror (if present) [m]", labelWidth=260, valueType=float, orientation="horizontal")

        oasysgui.lineEdit(box, self, "k", "Line Density (0th coeff.) [Lines/m]", labelWidth=260, valueType=float, orientation="horizontal")

        gui.separator(box)

        box_2 = oasysgui.widgetBox(tab_step_1, "Grating Design Parameters", orientation="vertical")

        gui.comboBox(box_2, self, "units_in_use", label="Units in use", labelWidth=260,
                     items=["eV", "Angstroms"],
                     callback=self.set_UnitsInUse, sendSelectedValue=False, orientation="horizontal")

        self.autosetting_box_units_1 = oasysgui.widgetBox(box_2, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.autosetting_box_units_1, self, "photon_energy", "Photon energy [eV]", labelWidth=260, valueType=float, orientation="horizontal")

        self.autosetting_box_units_2 = oasysgui.widgetBox(box_2, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.autosetting_box_units_2, self, "photon_wavelength", "Wavelength [Å]", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_UnitsInUse()

        oasysgui.lineEdit(box_2, self, "c", "C factor for optimized energy", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(box_2, self, "grating_diffraction_order", "Diffraction Order (- for inside orders)", labelWidth=260, valueType=int, orientation="horizontal")

        ##################################


        box_3 = oasysgui.widgetBox(tab_step_2, "Ray-Tracing Parameter", orientation="vertical")

        gui.comboBox(box_3, self, "new_units_in_use", label="Units in use", labelWidth=260,
                     items=["eV", "Angstroms"],
                     callback=self.set_UnitsInUse2, sendSelectedValue=False, orientation="horizontal")

        self.autosetting_box_units_3 = oasysgui.widgetBox(box_3, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.autosetting_box_units_3, self, "new_photon_energy", "New photon energy [eV]", labelWidth=260, valueType=float, orientation="horizontal")

        self.autosetting_box_units_4 = oasysgui.widgetBox(box_3, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.autosetting_box_units_4, self, "new_photon_wavelength", "New wavelength [Å]", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_UnitsInUse2()


        gui.comboBox(box_3, self, "new_c_flag", label="C factor for angles calculation", labelWidth=260,
                     items=["the same as for line density", "new one"],
                     callback=self.set_CfactorNew, sendSelectedValue=False, orientation="horizontal")

        self.c_box_new = oasysgui.widgetBox(box_3, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.c_box_new, self, "new_c_value", "new C for angles calculation", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_CfactorNew()


        tabs_out = oasysgui.tabWidget(self.mainArea)

        tab_out_1 = oasysgui.createTabPage(tabs_out, "Calculation Results")
        tab_out_2 = oasysgui.createTabPage(tabs_out, "Output")

        figure_box_1 = oasysgui.widgetBox(tab_out_1, "", addSpace=True, orientation="horizontal")

        label = QLabel("")
        label.setPixmap(QPixmap(self.image_path))

        figure_box_1.layout().addWidget(label)

        output_box = oasysgui.widgetBox(tab_out_1, "", addSpace=True, orientation="horizontal")

        output_box_1 = oasysgui.widgetBox(output_box, "Design Ouput", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(output_box_1, self, "design_alpha", "Alpha [deg]", labelWidth=220, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(output_box_1, self, "design_beta", "Beta [deg]", labelWidth=220, valueType=float, orientation="horizontal")
        gui.separator(output_box_1)
        oasysgui.lineEdit(output_box_1, self, "shadow_coeff_0", "Line Density 0-coeff. [Lines/m]", labelWidth=220, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(output_box_1, self, "shadow_coeff_1", "Line Density 1-coeff. [Lines/m\u00b2]", labelWidth=220, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(output_box_1, self, "shadow_coeff_2", "Line Density 2-coeff. [Lines/m\u00b3]", labelWidth=220, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(output_box_1, self, "shadow_coeff_3", "Line Density 3-coeff. [Lines/m\u00b4]", labelWidth=220, valueType=float, orientation="horizontal")

        output_box_2 = oasysgui.widgetBox(output_box, "Ray-Tracing Ouput", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(output_box_2, self, "raytracing_alpha", "Alpha [deg]", labelWidth=220, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(output_box_2, self, "raytracing_beta", "Beta [deg]", labelWidth=220, valueType=float, orientation="horizontal")
        gui.separator(output_box_2)
        oasysgui.lineEdit(output_box_2, self, "d_source_to_mirror", "Source to Mirror distance [m]", labelWidth=230, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(output_box_2, self, "d_source_plane_to_mirror", "Source Plane to Mirror distance [m]", labelWidth=230, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(output_box_2, self, "d_mirror_to_grating", "Mirror to Grating distance [m]", labelWidth=230, valueType=float, orientation="horizontal")


        self.shadow_output = oasysgui.textArea()

        out_box = oasysgui.widgetBox(tab_out_2, "System Output", addSpace=True, orientation="horizontal", height=400)
        out_box.layout().addWidget(self.shadow_output)

        gui.rubber(self.controlArea)

    def set_UnitsInUse(self):
        self.autosetting_box_units_1.setVisible(self.units_in_use == 0)
        self.autosetting_box_units_2.setVisible(self.units_in_use == 1)

    def set_UnitsInUse2(self):
        self.autosetting_box_units_3.setVisible(self.new_units_in_use == 0)
        self.autosetting_box_units_4.setVisible(self.new_units_in_use == 1)

    def set_CfactorNew(self):
        self.c_box_new.setVisible(self.new_c_flag == 1)

    @classmethod
    def calculate_vls_parameters(cls,
                                 m=-1,
                                 wavelength=1e-10,
                                 newwavelength=1e-10,
                                 k=0.0,
                                 c=1.0,
                                 r_a=0.0,
                                 r_b=0.0,
                                 h=0.0,
                                 new_c_flag=0,
                                 new_c_value=1.0,
                                 last_element_distance=0.0,
                                 verbose=1):

        sin_alpha = (-m * k * wavelength / (c ** 2 - 1)) + \
                    numpy.sqrt(1 + (m * m * c * c * k * k * wavelength * wavelength) / (
                                (c ** 2 - 1) ** 2))
        alpha = numpy.arcsin(sin_alpha)
        beta = numpy.arcsin(sin_alpha - m * k * wavelength)

        design_alpha = numpy.degrees(alpha)
        design_beta  = numpy.degrees(beta)

        if verbose:
            print("####################################################")
            print("# DESIGN PHASE")
            print("####################################################\n")

            print("Photon Wavelength [m]:", wavelength)
            print("Design ALPHA [deg]    :", round(design_alpha, 3))
            print("Design BETA [deg]     :", round(design_beta, 3))

        b2 = (((numpy.cos(alpha)**2) / r_a) + ((numpy.cos(beta)**2) / r_b)) / (-2 * m * k * wavelength)
        b3 = ((numpy.sin(alpha) * numpy.cos(alpha)**2) / r_a**2 - \
                  (numpy.sin(beta) * numpy.cos(beta)**2) / r_b**2) / (-2 * m * k * wavelength)
        b4 = (((4 * numpy.sin(alpha)**2 - numpy.cos(alpha)**2) * numpy.cos(alpha)**2) / r_a**3 + \
                  ((4 * numpy.sin(beta)**2 - numpy.cos(beta)**2) * numpy.cos(beta)**2) / r_b**3) / (-8 * m * k * wavelength)
        if verbose:
            print("\nb2:", b2)
            print("b3:",   b3)
            print("b4:",   b4)

        shadow_coeff_0 = k
        shadow_coeff_1 = -2 * k * b2
        shadow_coeff_2 = 3 * k * b3
        shadow_coeff_3 = -4 * k * b4

        if verbose:
            print("\nshadow_coeff_0:", round(shadow_coeff_0, 8))
            print("shadow_coeff_1:",   round(shadow_coeff_1, 8))
            print("shadow_coeff_2:",   round(shadow_coeff_2, 8))
            print("shadow_coeff_3:",   round(shadow_coeff_3, 8))

        ############################################
        #
        # 1 - in case of mirror recalculate real ray tracing distance (r_a') from initial r_a and vertical distance
        #     between grating and mirror (h)
        #

        gamma = (alpha + beta) / 2

        d_source_to_mirror = r_a - (h / numpy.abs(numpy.tan(numpy.pi - 2 * gamma)))
        d_mirror_to_grating = h / numpy.abs(numpy.sin(numpy.pi - 2 * gamma))

        r_a_first = d_source_to_mirror + d_mirror_to_grating

        if verbose:
            print("\ngamma [deg]                 :", numpy.degrees(gamma), "deg")
            print("Source to Mirror distance [m] :", d_source_to_mirror)
            print("Mirror to Grating distance [m]:", d_mirror_to_grating)
            print("R_a' [m]                      :", r_a_first)

        ############################################

        r = r_b / r_a_first

        A0 = k * newwavelength
        A2 = k * newwavelength * r_b * b2

        new_c_num = 2 * A2 + 4 * (A2 / A0)**2 + \
                    (4 + 2 * A2 - A0**2) * r - \
                    4 * (A2 / A0) * numpy.sqrt((1 + r)**2 + 2 * A2 * (1 + r) - r * A0**2)

        new_c_den = -4 + A0**2 - 4 * A2 + 4 * (A2 / A0)**2

        new_c = numpy.sqrt(new_c_num / new_c_den)

        if new_c_flag == 1:
            new_c = new_c_value

        new_sin_alpha = (-m * k * newwavelength / (new_c**2 - 1)) + \
                    numpy.sqrt(1 + (m * m * new_c * new_c * k * k * newwavelength * newwavelength) / ((new_c**2 - 1)**2))

        new_alpha =  numpy.arcsin(new_sin_alpha)
        new_beta  = numpy.arcsin(new_sin_alpha-m * k * newwavelength)

        raytracing_alpha = numpy.degrees(new_alpha)
        raytracing_beta  = numpy.degrees(new_beta)

        if verbose:

            #_new_beta = numpy.arccos(new_c*numpy.cos(new_alpha))

            print("####################################################")
            print("# RAY-TRACING PHASE")
            print("####################################################\n")

            print("Ray-Tracing Wavelength [m]:", newwavelength)
            print("Ray-Tracing Wavelength [A]:", newwavelength * 1e10)
            print("Ray-Tracing C         :", new_c)
            print("Ray-Tracing ALPHA     :", round(raytracing_alpha, 6), "deg")
            print("Ray-Tracing BETA      :", round(raytracing_beta, 6), "deg")

        gamma = (new_alpha + new_beta) / 2

        d_source_to_mirror  = r_a - (h / numpy.abs(numpy.tan(numpy.pi - 2 * gamma)))
        d_mirror_to_grating = h / numpy.abs(numpy.sin(numpy.pi - 2 * gamma))

        r_a_first = d_source_to_mirror + d_mirror_to_grating

        d_source_plane_to_mirror = d_source_to_mirror - last_element_distance
        if verbose:
            print("\ngamma [deg]                       :", numpy.degrees(gamma))
            print("Source to Mirror distance [m]       :", round(d_source_to_mirror, 3))
            print("Source Plane to Mirror distance [m] :", round(d_source_plane_to_mirror, 3))
            print("Mirror to Grating distance [m]      :", round(d_mirror_to_grating, 3))
            print("R_a' [m]                            :", r_a_first)

        return(
            shadow_coeff_0,
            shadow_coeff_1,
            shadow_coeff_2,
            shadow_coeff_3,
            d_source_to_mirror,
            d_source_plane_to_mirror,
            d_mirror_to_grating,
            design_alpha,
            design_beta,
            raytracing_alpha,
            raytracing_beta)




    def compute(self):
        try:
            self.shadow_output.setText("")

            sys.stdout = EmittingStream(textWritten=self.writeStdOut)

            self.checkFields()

            # m = -self.grating_diffraction_order
            #
            # if self.units_in_use == 0:
            #     wavelength =  ShadowPhysics.getWavelengthFromEnergy(self.photon_energy) * 1e-10
            # elif self.units_in_use == 1:
            #     wavelength = self.photon_wavelength * 1e-10
            #
            # sin_alpha = (-m * self.k * wavelength / (self.c**2 - 1)) + \
            #             numpy.sqrt(1 + (m * m * self.c * self.c * self.k * self.k * wavelength * wavelength) / ((self.c**2 - 1)**2))
            #
            # alpha = numpy.arcsin(sin_alpha)
            # beta = numpy.arcsin(sin_alpha-m * self.k * wavelength)
            #
            # self.design_alpha =  round(numpy.degrees(alpha), 3)
            # self.design_beta = round(numpy.degrees(beta), 3)
            # #_beta = numpy.arccos(self.c*numpy.cos(alpha))
            #
            # print("####################################################")
            # print("# DESIGN PHASE")
            # print("####################################################\n")
            #
            # print("Photon Wavelength [m]:", wavelength)
            # print("Design ALPHA [deg]    :", self.design_alpha)
            # print("Design BETA [deg]     :", self.design_beta)
            #
            # self.b2 = (((numpy.cos(alpha)**2) / self.r_a) + ((numpy.cos(beta)**2) / self.r_b)) / (-2 * m * self.k * wavelength)
            # self.b3 = ((numpy.sin(alpha) * numpy.cos(alpha)**2) / self.r_a**2 - \
            #           (numpy.sin(beta) * numpy.cos(beta)**2) / self.r_b**2) / (-2 * m * self.k * wavelength)
            # self.b4 = (((4 * numpy.sin(alpha)**2 - numpy.cos(alpha)**2) * numpy.cos(alpha)**2) / self.r_a**3 + \
            #           ((4 * numpy.sin(beta)**2 - numpy.cos(beta)**2) * numpy.cos(beta)**2) / self.r_b**3) / (-8 * m * self.k * wavelength)
            #
            # print("\nb2:", self.b2)
            # print("b3:", self.b3)
            # print("b4:", self.b4)
            #
            # self.shadow_coeff_0 = round(self.k, 8)
            # self.shadow_coeff_1 = round(-2 * self.k * self.b2, 8)
            # self.shadow_coeff_2 = round(3 * self.k * self.b3, 8)
            # self.shadow_coeff_3 = round(-4 * self.k * self.b4, 8)
            #
            # print("\nshadow_coeff_0:", self.shadow_coeff_0)
            # print("shadow_coeff_1:", self.shadow_coeff_1)
            # print("shadow_coeff_2:", self.shadow_coeff_2)
            # print("shadow_coeff_3:", self.shadow_coeff_3)
            #
            # ############################################
            # #
            # # 1 - in case of mirror recalculate real ray tracing distance (r_a') from initial r_a and vertical distance
            # #     between grating and mirror (h)
            # #
            #
            # gamma = (alpha + beta) / 2
            #
            # d_source_to_mirror = self.r_a - (self.h / numpy.abs(numpy.tan(numpy.pi - 2 * gamma)))
            # d_mirror_to_grating = self.h / numpy.abs(numpy.sin(numpy.pi - 2 * gamma))
            #
            # r_a_first = d_source_to_mirror + d_mirror_to_grating
            #
            # print("\ngamma [deg]                 :", numpy.degrees(gamma), "deg")
            # print("Source to Mirror distance [m] :", d_source_to_mirror)
            # print("Mirror to Grating distance [m]:", d_mirror_to_grating)
            # print("R_a' [m]                      :", r_a_first)
            #
            # ############################################
            #
            # if self.new_units_in_use == 0:
            #     newwavelength =  ShadowPhysics.getWavelengthFromEnergy(self.new_photon_energy) * 1e-10
            # elif self.new_units_in_use == 1:
            #     newwavelength = self.new_photon_wavelength * 1e-10
            #
            # r = self.r_b / r_a_first
            #
            # A0 = self.k * newwavelength
            # A2 = self.k * newwavelength * self.r_b * self.b2
            #
            # new_c_num = 2 * A2 + 4 * (A2 / A0)**2 + \
            #             (4 + 2 * A2 - A0**2) * r - \
            #             4 * (A2 / A0) * numpy.sqrt((1 + r)**2 + 2 * A2 * (1 + r) - r * A0**2)
            #
            # new_c_den = -4 + A0**2 - 4 * A2 + 4 * (A2 / A0)**2
            #
            # new_c = numpy.sqrt(new_c_num / new_c_den)
            #
            # if self.new_c_flag == 1:
            #     new_c = self.new_c_value
            #
            # new_sin_alpha = (-m * self.k*newwavelength / (new_c**2 - 1)) + \
            #             numpy.sqrt(1 + (m * m * new_c * new_c * self.k * self.k * newwavelength * newwavelength) / ((new_c**2 - 1)**2))
            #
            # new_alpha =  numpy.arcsin(new_sin_alpha)
            # new_beta  = numpy.arcsin(new_sin_alpha-m*self.k*newwavelength)
            #
            # self.raytracing_alpha = round(numpy.degrees(new_alpha), 6)
            # self.raytracing_beta  = round(numpy.degrees(new_beta), 6)
            # #_new_beta = numpy.arccos(new_c*numpy.cos(new_alpha))
            #
            # print("####################################################")
            # print("# RAY-TRACING PHASE")
            # print("####################################################\n")
            #
            # print("Ray-Tracing Wavelength [m]:", newwavelength)
            # print("Ray-Tracing Wavelength [A]:", newwavelength * 1e10)
            # print("Ray-Tracing C         :", new_c)
            # print("Ray-Tracing ALPHA     :", self.raytracing_alpha, "deg")
            # print("Ray-Tracing BETA      :", self.raytracing_beta, "deg")
            #
            # gamma = (new_alpha + new_beta)/2
            #
            # self.d_source_to_mirror  = self.r_a - (self.h / numpy.abs(numpy.tan(numpy.pi - 2 * gamma)))
            # self.d_mirror_to_grating = self.h / numpy.abs(numpy.sin(numpy.pi - 2 * gamma))
            #
            # r_a_first = self.d_source_to_mirror + self.d_mirror_to_grating
            #
            # self.d_source_to_mirror  = round(self.d_source_to_mirror , 3)
            # self.d_source_plane_to_mirror = round(self.d_source_to_mirror - self.last_element_distance, 3)
            # self.d_mirror_to_grating = round(self.d_mirror_to_grating, 3)
            #
            # print("\ngamma [deg]                       :", numpy.degrees(gamma))
            # print("Source to Mirror distance [m]       :", self.d_source_to_mirror)
            # print("Source Plane to Mirror distance [m] :", self.d_source_plane_to_mirror)
            # print("Mirror to Grating distance [m]      :", self.d_mirror_to_grating)
            # print("R_a' [m]                            :", r_a_first)


            m = -self.grating_diffraction_order

            if self.units_in_use == 0:
                wavelength =  ShadowPhysics.getWavelengthFromEnergy(self.photon_energy) * 1e-10
            elif self.units_in_use == 1:
                wavelength = self.photon_wavelength * 1e-10

            if self.new_units_in_use == 0:
                newwavelength =  ShadowPhysics.getWavelengthFromEnergy(self.new_photon_energy) * 1e-10
            elif self.new_units_in_use == 1:
                newwavelength = self.new_photon_wavelength * 1e-10


            shadow_coeff_0, shadow_coeff_1, shadow_coeff_2, shadow_coeff_3, \
            d_source_to_mirror, d_source_plane_to_mirror, d_mirror_to_grating, \
            design_alpha, design_beta, raytracing_alpha, raytracing_beta = \
                self.calculate_vls_parameters(
                    m=m,
                    wavelength=wavelength,
                    newwavelength=newwavelength,
                    k=self.k,
                    c=self.c,
                    r_a=self.r_a,
                    r_b=self.r_b,
                    h=self.h,
                    new_c_flag=self.new_c_flag,
                    new_c_value=self.new_c_value,
                    last_element_distance=self.last_element_distance,
                    verbose=1)

            self.shadow_coeff_0           = round(shadow_coeff_0, 3)
            self.shadow_coeff_1           = round(shadow_coeff_1, 3)
            self.shadow_coeff_2           = round(shadow_coeff_2, 3)
            self.shadow_coeff_3           = round(shadow_coeff_3, 3)
            self.d_source_plane_to_mirror = round(d_source_plane_to_mirror, 6)
            self.d_source_to_mirror       = round(d_source_to_mirror, 6)
            self.d_mirror_to_grating      = round(d_mirror_to_grating, 6)
            self.design_alpha             = round(design_alpha, 6)
            self.design_beta              = round(design_beta, 6)
            self.raytracing_alpha         = round(raytracing_alpha, 6)
            self.raytracing_beta          = round(raytracing_beta, 6)


            self.send("VLS-PGM PreProcessor Data", VlsPgmPreProcessorData(shadow_coeff_0=self.shadow_coeff_0,
                                                                          shadow_coeff_1=self.shadow_coeff_1,
                                                                          shadow_coeff_2=self.shadow_coeff_2,
                                                                          shadow_coeff_3=self.shadow_coeff_3,
                                                                          d_source_plane_to_mirror=self.d_source_plane_to_mirror,
                                                                          d_mirror_to_grating=self.d_mirror_to_grating,
                                                                          d_grating_to_exit_slits=self.r_b,
                                                                          alpha=self.raytracing_alpha,
                                                                          beta=self.raytracing_beta))

        except Exception as exception:
            QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)
            if self.IS_DEVELOP: raise exception

    def checkFields(self):
        self.r_a = congruence.checkPositiveNumber(self.r_a, "Distance Source-Grating")
        self.r_b = congruence.checkPositiveNumber(self.r_b, "Distance Grating-Exit Slits")
        self.last_element_distance = congruence.checkPositiveNumber(self.last_element_distance, "Distance Source-Last Image Plane before Mirror")
        congruence.checkLessOrEqualThan(self.last_element_distance, self.r_a, "Distance Source-Last Image Plane before Mirror", "Distance Source-Grating")
        self.k = congruence.checkStrictlyPositiveNumber(self.k, "Line Density")

        if self.units_in_use == 0:
            self.photon_energy = congruence.checkPositiveNumber(self.photon_energy, "Photon Energy")
        elif self.units_in_use == 1:
            self.photon_wavelength = congruence.checkPositiveNumber(self.photon_wavelength, "Photon Wavelength")



    def defaults(self):
         self.resetSettings()

    def writeStdOut(self, text):
        cursor = self.shadow_output.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        self.shadow_output.setTextCursor(cursor)
        self.shadow_output.ensureCursorVisible()

if __name__ == "__main__":
    app = QApplication(sys.argv)
    w = OWVlsPgmCoefficientsCalculator()
    w.show()
    app.exec()
    w.saveSettings()



import sys, os
import time
import numpy

from PyQt5.QtGui import QTextCursor
from PyQt5 import QtGui, QtWidgets
from PyQt5.QtWidgets import QMessageBox

from orangewidget import gui
from orangewidget.settings import Setting
from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence
from oasys2.widget.gui import ConfirmDialog, MessageDialog, selectSaveFileFromDialog

from oasys2.widget.util.widget_util import EmittingStream

from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence, ShadowPlot
from orangecontrib.shadow4.widgets.gui.ow_automatic_element import AutomaticElement
from orangecontrib.shadow4.util.python_script import PythonScript

from shadow4.beam.s4_beam import S4Beam

from silx.gui.plot.ScatterView import ScatterView
from silx.gui.colors import Colormap

from oasys2.widget.util.widget_util import EmittingStream


try:
    import OpenGL
    has_opengl = True
except:
    has_opengl = False

class PlotScatter(AutomaticElement):
    name = "Plot Scatter"
    description = "Display Data Tools: Plot XY"
    icon = "icons/plot_scatter.png"
    maintainer = "Manuel Sanchez del Rio"
    maintainer_email = "srio(@at@)esrf.eu"
    priority = 1.3
    category = "Display Data Tools"
    keywords = ["data", "file", "load", "read"]

    inputs = [("Shadow Data", ShadowData, "set_shadow_data"),
              ("Shadow Data (color)", ShadowData, "set_shadow_data_for_color")]

    IMAGE_WIDTH = 878
    IMAGE_HEIGHT = 635

    want_main_area = 1

    plot_canvas = None

    image_plane = Setting(0)
    image_plane_new_position = Setting(10.0)

    x_column_index = Setting(0)
    x_range = Setting(0)
    x_range_min = Setting(0.0)
    x_range_max = Setting(0.0)

    y_column_index = Setting(2)
    y_range = Setting(0)
    y_range_min = Setting(0.0)
    y_range_max = Setting(0.0)

    weight_transparency = Setting(0)

    color_source = Setting(0) # 0 = same beam , 1 = specific beam
    color_column = Setting(23)
    weight_transparency = Setting(0)

    rays = Setting(1)
    title = Setting("")

    conversion_active = Setting(1)

    if has_opengl:
        backend = 1
    else:
        backend = 0

    input_data = None
    input_data_color = None

    def __init__(self):
        super().__init__()

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        gui.button(button_box, self, "Refresh", callback=self.plot_results, height=45)

        gui.separator(self.controlArea, 10)

        self.tabs_setting = oasysgui.tabWidget(self.controlArea)
        self.tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH - 5)

        # graph tab
        tab_set = oasysgui.createTabPage(self.tabs_setting, "Plot Settings")

        ################
        screen_box = oasysgui.widgetBox(tab_set, "Screen Position Settings", addSpace=True, orientation="vertical",)

        self.image_plane_combo = gui.comboBox(screen_box, self, "image_plane", label="Position of the Image",
                                              items=["On Image Plane", "Retraced"], labelWidth=260,
                                              callback=self.set_visibility, sendSelectedValue=False,
                                              orientation="horizontal")

        self.image_plane_box = oasysgui.widgetBox(screen_box, "", addSpace=False, orientation="vertical",) # height=50)

        oasysgui.lineEdit(self.image_plane_box, self, "image_plane_new_position", "New (relative) position [m]",
                          labelWidth=220, valueType=float, orientation="horizontal")

        ################
        general_box = oasysgui.widgetBox(tab_set, "Variables Settings", addSpace=True, orientation="vertical",)


        #######
        self.x_column = gui.comboBox(general_box, self, "x_column_index", label="H Column", labelWidth=70,
                                     items=S4Beam.column_names_with_column_number(),
                                     sendSelectedValue=False, orientation="horizontal",
                                     callback=self.set_x_column_index)

        gui.comboBox(general_box, self, "x_range", label="H Range", labelWidth=250,
                     items=["<Default>", "Set.."],
                     callback=self.set_visibility, sendSelectedValue=False, orientation="horizontal")

        self.x_range_box = oasysgui.widgetBox(general_box, "", addSpace=True, orientation="vertical",) # height=120)

        self.le_x_range_min = oasysgui.lineEdit(self.x_range_box, self, "x_range_min", "H min", labelWidth=220,
                                                valueType=float, orientation="horizontal")
        self.le_x_range_max = oasysgui.lineEdit(self.x_range_box, self, "x_range_max", "H max", labelWidth=220,
                                                valueType=float, orientation="horizontal")

        #######
        self.y_column = gui.comboBox(general_box, self, "y_column_index", label="V Column", labelWidth=70,
                                     items=S4Beam.column_names_with_column_number(),
                                     sendSelectedValue=False, orientation="horizontal",
                                     callback=self.set_y_column_index)

        gui.comboBox(general_box, self, "y_range", label="V Range", labelWidth=250,
                     items=["<Default>", "Set.."],
                     callback=self.set_visibility, sendSelectedValue=False, orientation="horizontal")

        self.y_range_box = oasysgui.widgetBox(general_box, "", addSpace=True, orientation="vertical",) # height=100)

        self.le_y_range_min = oasysgui.lineEdit(self.y_range_box, self, "y_range_min", "V min", labelWidth=220,
                                                valueType=float, orientation="horizontal")
        self.le_y_range_max = oasysgui.lineEdit(self.y_range_box, self, "y_range_max", "V max", labelWidth=220,
                                                valueType=float, orientation="horizontal")

        #######
        gui.comboBox(general_box, self, "color_source", label="Color source", labelWidth=100,
                     items=["The same beam", "Specific beam"],sendSelectedValue=False, orientation="horizontal")

        ######
        col_names = S4Beam.column_names_with_column_number()
        col_names.insert(0, "0: No Weight")
        self.weight_column = gui.comboBox(general_box, self, "color_column", label="Color", labelWidth=70,
                                          items=col_names,
                                          sendSelectedValue=False, orientation="horizontal")


        gui.comboBox(general_box, self, "weight_transparency", label="Transparency (Weight=col23)", labelWidth=250,
                                         items=["No","Yes"],
                                         sendSelectedValue=False, orientation="horizontal")

        gui.comboBox(general_box, self, "rays", label="Rays", labelWidth=250,
                     items=["All rays",
                            "Good Only",
                            "Lost Only"],
                     sendSelectedValue=False, orientation="horizontal")

        gui.comboBox(general_box, self, "conversion_active", label="Is U.M. conversion active", labelWidth=250,
                     items=["No", "Yes"],
                     sendSelectedValue=False, orientation="horizontal", callback=self.set_is_conversion_active)

        gui.comboBox(general_box, self, "backend", label="render backend", labelWidth=250,
                                         items=["matplotlib", "gl"],
                                         sendSelectedValue=False, orientation="horizontal")

        self.main_tabs = oasysgui.tabWidget(self.mainArea)
        plot_tab = oasysgui.createTabPage(self.main_tabs, "Plots")
        out_tab = oasysgui.createTabPage(self.main_tabs, "Output")
        script_tab = oasysgui.createTabPage(self.main_tabs, "Script")

        self.image_box = gui.widgetBox(plot_tab, "Plot Result", addSpace=True, orientation="vertical")
        self.image_box.setFixedHeight(self.IMAGE_HEIGHT)
        self.image_box.setFixedWidth(self.IMAGE_WIDTH)

        self.shadow4_script = PythonScript()
        self.shadow4_script.code_area.setFixedHeight(400)
        script_box = gui.widgetBox(script_tab, "Python script", addSpace=True, orientation="horizontal")
        script_box.layout().addWidget(self.shadow4_script)

        self.shadow_output = oasysgui.textArea(height=580, width=800)

        out_box = gui.widgetBox(out_tab, "System Output", addSpace=True, orientation="horizontal")
        out_box.layout().addWidget(self.shadow_output)

        self.set_visibility()

    def set_shadow_data(self, shadow_data: ShadowData):
        if ShadowCongruence.check_empty_data(shadow_data):
            if ShadowCongruence.check_empty_beam(shadow_data.beam):
                self.input_data = shadow_data
                if self.is_automatic_run: self.plot_results()
            else:
                MessageDialog.message(self, "Data not displayable: bad content", "Error", "critical")

    def set_shadow_data_for_color(self, shadow_data: ShadowData):
        if ShadowCongruence.check_empty_data(shadow_data):
            if ShadowCongruence.check_empty_beam(shadow_data.beam):
                self.input_data_color = shadow_data
                if self.is_automatic_run: self.plot_results()
            else:
                MessageDialog.message(self, "Data not displayable: bad content", "Error", "critical")

    def writeStdOut(self, text):
        cursor = self.shadow_output.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        self.shadow_output.setTextCursor(cursor)
        self.shadow_output.ensureCursorVisible()

    def retrace_beam(self, new_shadow_beam: S4Beam, dist):
        new_shadow_beam.retrace(dist)

    def set_visibility(self):
        self.x_range_box.setVisible(False)
        self.y_range_box.setVisible(False)
        self.image_plane_box.setVisible(False)

        if self.x_range == 1:
            self.x_range_box.setVisible(True)

        if self.y_range == 1:
            self.y_range_box.setVisible(True)

        self.__change_labels(dir='x')
        self.__change_labels(dir='y')

        if self.image_plane == 1:
            self.image_plane_box.setVisible(True)

    def set_is_conversion_active(self):
        self.__change_labels(dir='b')

    def set_script(self, x_range, y_range):

        # script
        try:
            beamline = self.input_data.beamline.duplicate()

            script = beamline.to_python_code(partial_code=1) # onlu source
            indented_script = '\n'.join('    ' + line for line in script.splitlines())
            final_script = "def run_source():\n"
            final_script += indented_script
            final_script += "\n    return beam"
            final_script += "\n\n"

            script = beamline.to_python_code(partial_code=2) # only beamline
            indented_script = '\n'.join('    ' + line for line in script.splitlines())
            final_script += "def run_beamline(beam):\n"
            final_script += indented_script
            final_script += "\n    return beam, footprint"
            final_script += "\n\n"

            # TODO: add script for color
            if self.color_source == 1:
                beamline_color = self.input_data_color.beamline.duplicate()
                script = beamline_color.to_python_code(partial_code=2) # only beamline
                indented_script = '\n'.join('    ' + line for line in script.splitlines())
                final_script += "def run_beamline_for_color(beam):\n"
                final_script += indented_script
                final_script += "\n    return beam"
                final_script += "\n\n"

            if self.image_plane > 0:
                retrace = "beam.retrace(%f)" % self.image_plane_new_position
            else:
                retrace = ""

            if self.color_source == 0:
                color_same_beam = "(the same beam)"
                color_code = "beam_color = beam"
            else:
                color_same_beam = "(from different beam)"
                color_code = "beam_color = run_beamline_for_color(beam_source)"

            dict = {"var_x": 1 + self.x_column_index,
                    "var_y": 1 + self.y_column_index,
                    "x_range_min": x_range[0],
                    "x_range_max": x_range[1],
                    "y_range_min": y_range[0],
                    "y_range_max": y_range[1],
                    "nolost": self.rays,
                    "color_column": self.color_column,
                    "color_code": color_code,
                    "weight_transparency": self.weight_transparency,
                    "retrace": retrace,
                    "color_code": color_code,
                    "color_same_beam": color_same_beam,
                    }

            script_template = """#
#
# main
#
import matplotlib.pyplot as plt

# WARNING: No incremental or cumulated result allowed!!"
beam_source = run_source()
beam, footprint = run_beamline(beam_source)
{retrace}

{color_code}

weight_transparency = {weight_transparency} # flag to use transparency

x = beam.get_column({var_x}, nolost={nolost}) # H 
y = beam.get_column({var_y}, nolost={nolost}) # V
t = beam.get_column(23, nolost={nolost}) # for transparency
f = beam.get_column(10, nolost={nolost}) # lost ray flag
i = beam.get_column(12, nolost={nolost})  # index ray flag

colors = beam_color.get_column({color_column}, nolost=0)
colors = colors[i.astype(int) - 1]

if weight_transparency == 0: t = t * 0 + 1
alpha_vals = t / t.max()  # normalize to [0, 1]

# Create scatter plot (no border, filled with color)
sc = plt.scatter(
    x, y,
    c=colors,
    marker='o',
    cmap='viridis',
    edgecolors='none'  # removes circle border
)

# Apply individual alpha values
facecolors = sc.get_facecolors()
for i in range(len(facecolors)):
    facecolors[i][-1] = alpha_vals[i]
sc.set_facecolors(facecolors)

# Labels
plt.xlabel('column {var_x}')
plt.ylabel('column {var_y}')

# Colorbar
plt.colorbar(label='column {color_column} {color_same_beam}')

# limitx
plt.xlim({x_range_min}, {x_range_max})
plt.ylim({y_range_min}, {y_range_max})

plt.title('Scatter Plot, transparency=%d' % (weight_transparency))
plt.grid(True)
plt.show()

"""

            final_script += script_template.format_map(dict)

            self.shadow4_script.set_code(final_script)
        except:
            final_script += "\n\n\n# cannot retrieve beamline data from shadow_data"

        self.shadow4_script.set_code(final_script)


    #######################

    def plot_results(self):

        try:
            plotted = False

            sys.stdout = EmittingStream(textWritten=self.writeStdOut)

            if ShadowCongruence.check_empty_data(self.input_data):
                ShadowPlot.set_conversion_active(self.is_conversion_active())

                x, y, c, auto_x_title, auto_y_title, xum, yum = self.get_titles()

                print(">>>>", x, y, c, auto_x_title, auto_y_title, xum, yum)

                self.plot_scatter(x, y, c, title=self.title, xtitle=auto_x_title, ytitle=auto_y_title, xum=xum, yum=yum)

                plotted = True

            time.sleep(0.1)  # prevents a misterious dead lock in the Orange cycle when refreshing the histogram

            return plotted


            if True: # self.trace_shadow:
                grabber.stop()

                for row in grabber.ttyData:
                    self.writeStdOut(row)

            time.sleep(0.5)  # prevents a misterious dead lock in the Orange cycle when refreshing the histogram

            return plotted
        except Exception as exception:
            QtWidgets.QMessageBox.critical(self, "Error",
                                       str(exception),
                                       QtWidgets.QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

            return False

    def plot_scatter(self, var_x, var_y, var_color, title, xtitle, ytitle, xum, yum):
        beam_to_plot = self.input_data.beam

        if self.image_plane == 1:
            new_shadow_beam = beam_to_plot.duplicate()
            self.retrace_beam(new_shadow_beam, self.image_plane_new_position)
            beam_to_plot = new_shadow_beam

        xrange, yrange = self.get_ranges(beam_to_plot, var_x, var_y)

        self.set_script(x_range=xrange, y_range=yrange)

        self.replace_fig(beam_to_plot, var_x, var_y, var_color,
                         title, xtitle, ytitle,
                         xrange=xrange, yrange=yrange,
                         nolost=self.rays, xum=xum, yum=yum)



    def replace_fig(self, beam, var_x, var_y, var_color, title, xtitle, ytitle,
                    xrange=[0,0], yrange=[0,0], nolost=0, xum="", yum=""):

        if self.backend == 0:
            use_backend = 'matplotlib'
        elif self.backend == 1:
            if not has_opengl:
                QMessageBox.information(self, "Plot Scatter Information",
                        "It seems that PyOpenGL is not installed in your system." +
                        "\nInstall it to get much faster scatter plots, like:" +
                        "\n" + os.path.dirname(sys.executable) + os.sep + "pip install PyOpenGL",
                        QMessageBox.Ok)
                use_backend = 'matplotlib'
                self.backend = 0
            else:
                use_backend = 'gl'

        if self.plot_canvas is None:
            self.plot_canvas = ScatterView(backend=use_backend)
        else:
            self.image_box.layout().removeWidget(self.plot_canvas)
            self.plot_canvas = None
            self.plot_canvas = ScatterView(backend=use_backend)

        if self.color_column != 0:
            if self.color_source == 0:
                color_array = beam.get_column(var_color, nolost=nolost)
            else:
                if self.input_data_color is None:
                    raise Exception("Undefined specific beam for color")

                color_array = self.input_data_color.beam.get_column(var_color, nolost=False)
                if nolost == 0:
                    pass
                elif nolost == 1:
                    color_good_flags = beam.get_column(10, nolost=False)
                    color_array = color_array[numpy.where(color_good_flags >=0 )]
                elif nolost == 2:
                    color_good_flags = beam.get_column(10, nolost=False)
                    color_array = color_array[numpy.where(color_good_flags < 0 )]
        else:
            color_array = beam.get_column(1, nolost=nolost) * 0.0

        factor1 = ShadowPlot.get_factor(var_x)
        factor2 = ShadowPlot.get_factor(var_y)
        factorC = ShadowPlot.get_factor(var_color)


        if self.weight_transparency == 1:
            self.plot_canvas.setData(
                beam.get_column(var_x, nolost=nolost)*factor1,
                beam.get_column(var_y, nolost=nolost)*factor2,
                color_array*factorC,
                alpha = beam.get_column(23, nolost=nolost) )
        else:
            self.plot_canvas.setData(
                beam.get_column(var_x, nolost=nolost)*factor1,
                beam.get_column(var_y, nolost=nolost)*factor2,
                color_array*factorC,
                )

        self.plot_canvas.resetZoom()
        self.plot_canvas.setGraphTitle(title)
        self.plot_canvas.setColormap(Colormap('viridis'))

        ax = self.plot_canvas.getPlotWidget().getXAxis()
        if self.x_range == 1:
            ax.setLimits(self.x_range_min,self.x_range_max)
        ax.setLabel(xtitle)

        ay = self.plot_canvas.getPlotWidget().getYAxis()
        if self.y_range == 1:
            ay.setLimits(self.y_range_min,self.y_range_max)
        ay.setLabel(ytitle)

        self.image_box.layout().addWidget(self.plot_canvas)

    def set_x_column_index(self):
        self.__change_labels(dir='x')

    def set_y_column_index(self):
        self.__change_labels(dir='y')

    def __change_labels(self, dir='b'):
        def change_label(line_edit, index):
            label = line_edit.parent().layout().itemAt(0).widget()
            label_text = label.text()
            if label_text[-1] == "]":
                label_text = label_text.split(sep="[")[0]
            else:
                label_text += " "
            if index in [0, 1, 2] and self.is_conversion_active():
                label_text += "[\u03BCm]"
            elif index in [3, 4, 5] and self.is_conversion_active():
                label_text += "[\u03BCrad]"
            else:
                label_text += S4Beam.column_units()[index]
            label.setText(label_text)

        if self.x_range == 1 and dir in ['x', 'b']:
            change_label(self.le_x_range_min, self.x_column_index)
            change_label(self.le_x_range_max, self.x_column_index)
        if self.y_range == 1 and dir in ['y', 'b']:
            change_label(self.le_y_range_min, self.y_column_index)
            change_label(self.le_y_range_max, self.y_column_index)

    def get_ranges(self, beam_to_plot, var_x, var_y):
        factor1 = ShadowPlot.get_factor(var_x)
        factor2 = ShadowPlot.get_factor(var_y)

        if self.x_range == 1:
            congruence.checkLessThan(self.x_range_min, self.x_range_max, "X range min", "X range max")
            x_range = [self.x_range_min / factor1, self.x_range_max / factor1]
        else:
            x, y = beam_to_plot.get_columns((var_x, var_y), nolost=self.rays)
            x_max = x.max()
            x_min = x.min()
            if numpy.abs(x_max - x_min) < 1e-10:
                x_min -= 1e-10
                x_max -= 1e-10
            x_range = [x_min, x_max]

        if self.y_range == 1:
            congruence.checkLessThan(self.y_range_min, self.y_range_max, "Y range min", "Y range max")
            y_range = [self.y_range_min / factor2, self.y_range_max / factor2]
        else:
            x, y = beam_to_plot.get_columns((var_x, var_y), nolost=self.rays)
            y_max = y.max()
            y_min = y.min()
            if numpy.abs(y_max - y_min) < 1e-10:
                y_min -= 1e-10
                y_max -= 1e-10
            y_range = [y_min, y_max]

        return x_range, y_range

    def get_titles(self):
        xum = auto_x_title = self.x_column.currentText()
        yum = auto_y_title = self.y_column.currentText()

        self.title = S4Beam.column_short_names()[self.x_column_index] + "," + S4Beam.column_short_names()[
            self.y_column_index]

        def get_strings(um, auto_title, col, index):
            if col in [1, 2, 3] and self.is_conversion_active():
                um += " [\u03BCm]"
                auto_title += " [\u03BCm]" # " [\mu m]"
            elif col in [4, 5, 6] and self.is_conversion_active():
                um += " [\u03BCrad]"
                auto_title += " [\u03BCrad]" # " [\mu rad]"
            else:
                um += " " + S4Beam.column_units()[index]
                auto_title += " " + S4Beam.column_units()[index]

            return um, auto_title

        x = self.x_column_index + 1
        y = self.y_column_index + 1
        c = self.color_column
        xum, auto_x_title = get_strings(xum, auto_x_title, x, self.x_column_index)
        yum, auto_y_title = get_strings(yum, auto_y_title, y, self.y_column_index)

        return x, y, c, auto_x_title, auto_y_title, xum, yum

    def is_conversion_active(self):
        return self.conversion_active == 1


if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication, QMessageBox


    def get_beamline():
        from shadow4.beamline.s4_beamline import S4Beamline

        beamline = S4Beamline()

        # electron beam
        from shadow4.sources.s4_electron_beam import S4ElectronBeam
        electron_beam = S4ElectronBeam(energy_in_GeV=6, energy_spread=0.001, current=0.2)
        electron_beam.set_sigmas_all(sigma_x=3.01836e-05, sigma_y=4.36821e-06, sigma_xp=3.63641e-06,
                                     sigma_yp=1.37498e-06)

        # magnetic structure
        from shadow4.sources.undulator.s4_undulator_gaussian import S4UndulatorGaussian
        source = S4UndulatorGaussian(
            period_length=0.042,  # syned Undulator parameter (length in m)
            number_of_periods=38.571,  # syned Undulator parameter
            photon_energy=5000.0,  # Photon energy (in eV)
            delta_e=4.0,  # Photon energy width (in eV)
            ng_e=100,  # Photon energy scan number of points
            flag_emittance=1,  # when sampling rays: Use emittance (0=No, 1=Yes)
            flag_energy_spread=0,  # when sampling rays: Use e- energy spread (0=No, 1=Yes)
            harmonic_number=1,  # harmonic number
            flag_autoset_flux_central_cone=1,  # value to set the flux peak
            flux_central_cone=681709040139326.4,  # value to set the flux peak
        )

        # light source
        from shadow4.sources.undulator.s4_undulator_gaussian_light_source import S4UndulatorGaussianLightSource
        light_source = S4UndulatorGaussianLightSource(name='GaussianUndulator', electron_beam=electron_beam,
                                                      magnetic_structure=source, nrays=15000, seed=5676561)
        beam = light_source.get_beam()

        beamline.set_light_source(light_source)

        # optical element number XX
        from syned.beamline.shape import Rectangle
        boundary_shape = Rectangle(x_left=-0.001, x_right=0.001, y_bottom=-0.001, y_top=0.001)

        from shadow4.beamline.optical_elements.absorbers.s4_screen import S4Screen
        optical_element = S4Screen(name='Generic Beam Screen/Slit/Stopper/Attenuator', boundary_shape=boundary_shape,
                                   i_abs=0,  # 0=No, 1=prerefl file_abs, 2=xraylib, 3=dabax
                                   i_stop=0, thick=0, file_abs='<specify file name>', material='Au', density=19.3)

        from syned.beamline.element_coordinates import ElementCoordinates
        coordinates = ElementCoordinates(p=27.2, q=0, angle_radial=0, angle_azimuthal=0, angle_radial_out=3.141592654)
        from shadow4.beamline.optical_elements.absorbers.s4_screen import S4ScreenElement
        beamline_element = S4ScreenElement(optical_element=optical_element, coordinates=coordinates, input_beam=beam)

        beam, footprint = beamline_element.trace_beam()

        beamline.append_beamline_element(beamline_element)

        # optical element number XX
        boundary_shape = None

        from shadow4.beamline.optical_elements.mirrors.s4_plane_mirror import S4PlaneMirror
        optical_element = S4PlaneMirror(name='Plane Mirror', boundary_shape=boundary_shape,
                                        f_reflec=1, f_refl=5, file_refl='<none>', refraction_index=0.99999 + 0.001j,
                                        coating_material='Ni', coating_density=8.902, coating_roughness=0)

        from syned.beamline.element_coordinates import ElementCoordinates
        coordinates = ElementCoordinates(p=2.7, q=0, angle_radial=1.563796327, angle_azimuthal=1.570796327,
                                         angle_radial_out=1.563796327)
        movements = None
        from shadow4.beamline.optical_elements.mirrors.s4_plane_mirror import S4PlaneMirrorElement
        beamline_element = S4PlaneMirrorElement(optical_element=optical_element, coordinates=coordinates,
                                                movements=movements, input_beam=beam)

        beam, mirr = beamline_element.trace_beam()

        beamline.append_beamline_element(beamline_element)

        # test plot
        if 0:
            from srxraylib.plot.gol import plot_scatter
            plot_scatter(beam.get_photon_energy_eV(nolost=1), beam.get_column(23, nolost=1),
                         title='(Intensity,Photon Energy)', plot_histograms=0)
            plot_scatter(1e6 * beam.get_column(1, nolost=1), 1e6 * beam.get_column(3, nolost=1),
                         title='(X,Z) in microns')
        return beam, footprint, beamline
        ###############################


    beam, footprint, beamline = get_beamline()

    app = QApplication(sys.argv)
    w = PlotScatter()
    w.set_shadow_data(ShadowData(beam=beam, footprint=footprint, number_of_rays=0, beamline=beamline))
    w.set_shadow_data_for_color(ShadowData(beam=beam, footprint=footprint, number_of_rays=0, beamline=beamline))
    w.show()
    app.exec()




from PyQt5.QtGui import QPalette, QColor, QFont

from orangewidget import gui
from orangewidget import widget
from orangewidget.settings import Setting

from oasys2.widget import gui as oasysgui
from oasys2.widget.util import congruence

from orangecontrib.shadow4.widgets.gui.ow_generic_element import GenericElement

from shadow4.sources.s4_electron_beam import S4ElectronBeam

class OWElectronBeam(GenericElement):

    syned_file_name = Setting("Select *.json file")

    electron_energy_in_GeV = Setting(1.9)
    electron_energy_spread = Setting(0.000)
    ring_current           = Setting(0.4)

    moment_xx           = Setting(0.0)
    moment_xxp          = Setting(0.0)
    moment_xpxp         = Setting(0.0)
    moment_yy           = Setting(0.0)
    moment_yyp          = Setting(0.0)
    moment_ypyp         = Setting(0.0)

    electron_beam_size_h       = Setting(39e-6)
    electron_beam_divergence_h = Setting(31e-6)
    electron_beam_size_v       = Setting(39.2e-6)
    electron_beam_divergence_v = Setting(39.2e-6)

    electron_beam_emittance_h = Setting(0.0)
    electron_beam_emittance_v = Setting(0.0)
    electron_beam_beta_h = Setting(0.0)
    electron_beam_beta_v = Setting(0.0)
    electron_beam_alpha_h = Setting(0.0)
    electron_beam_alpha_v = Setting(0.0)
    # electron_beam_eta_h = Setting(0.0)
    # electron_beam_eta_v = Setting(0.0)
    # electron_beam_etap_h = Setting(0.0)
    # electron_beam_etap_v = Setting(0.0)

    type_of_properties = Setting(1)
    flag_energy_spread = Setting(0)

    def __init__(self, show_energy_spread=False):
        super().__init__(show_automatic_box=False, has_footprint=False)

        self.runaction = widget.OWAction("Run Shadow4/Source", self)
        self.runaction.triggered.connect(self.run_shadow4)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Run shadow4/source", callback=self.run_shadow4)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Fields", callback=self.call_reset_settings)
        font = QFont(button.font())
        font.setItalic(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Red'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)
        button.setFixedWidth(150)


        self.tabs_control_area = oasysgui.tabWidget(self.controlArea)
        self.tabs_control_area.setFixedHeight(self.TABS_AREA_HEIGHT)
        self.tabs_control_area.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        self.tab_electron_beam = oasysgui.createTabPage(self.tabs_control_area, "Electron Beam Setting")

        self.electron_beam_box = oasysgui.widgetBox(self.tab_electron_beam, "Electron Beam/Machine Parameters", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.electron_beam_box, self, "electron_energy_in_GeV", "Energy [GeV]",  tooltip="electron_energy_in_GeV", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.electron_beam_box, self, "ring_current", "Ring Current [A]",        tooltip="ring_current",           labelWidth=260, valueType=float, orientation="horizontal")

        if show_energy_spread:
            gui.comboBox(self.electron_beam_box, self, "flag_energy_spread", tooltip="flag_energy_spread", label="Energy Spread", labelWidth=350,
                     items=["No (zero)", "Yes"],
                     callback=self.set_TypeOfProperties,
                     sendSelectedValue=False, orientation="horizontal")
        else:
            self.flag_energy_spread = 0
        self.box_energy_spread = oasysgui.widgetBox(self.electron_beam_box, "", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(self.box_energy_spread, self, "electron_energy_spread", "Energy Spread DE/E", tooltip="electron_energy_spread", labelWidth=260, valueType=float, orientation="horizontal")


        gui.comboBox(self.electron_beam_box, self, "type_of_properties", tooltip="type_of_properties", label="Electron Beam Properties", labelWidth=350,
                     items=["From 2nd Moments", "From Size/Divergence", "From Twiss parameters","Zero emittance"],
                     callback=self.set_TypeOfProperties,
                     sendSelectedValue=False, orientation="horizontal")

        self.left_box_2_1 = oasysgui.widgetBox(self.electron_beam_box, "", addSpace=False, orientation="vertical", height=150)

        oasysgui.lineEdit(self.left_box_2_1, self, "moment_xx",   "<x x>   [m^2]",   tooltip="moment_xx",   labelWidth=160, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_xxp",  "<x x'>  [m.rad]", tooltip="moment_xxp",  labelWidth=160, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_xpxp", "<x' x'> [rad^2]", tooltip="moment_xpxp", labelWidth=160, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_yy",   "<y y>   [m^2]",   tooltip="moment_yy",   labelWidth=160, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_yyp",  "<y y'>  [m.rad]", tooltip="moment_yyp",  labelWidth=160, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_ypyp", "<y' y'> [rad^2]", tooltip="moment_ypyp", labelWidth=160, valueType=float, orientation="horizontal")


        self.left_box_2_2 = oasysgui.widgetBox(self.electron_beam_box, "", addSpace=False, orientation="vertical", height=150)

        oasysgui.lineEdit(self.left_box_2_2, self, "electron_beam_size_h",       "Horizontal Beam Size \u03c3x [m]",          tooltip="electron_beam_size_h", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_2, self, "electron_beam_size_v",       "Vertical Beam Size \u03c3y [m]",            tooltip="electron_beam_size_v",  labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_2, self, "electron_beam_divergence_h", "Horizontal Beam Divergence \u03c3'x [rad]", tooltip="electron_beam_divergence_h", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_2, self, "electron_beam_divergence_v", "Vertical Beam Divergence \u03c3'y [rad]",   tooltip="electron_beam_divergence_v", labelWidth=260, valueType=float, orientation="horizontal")

        self.left_box_2_3 = oasysgui.widgetBox(self.electron_beam_box, "", addSpace=False, orientation="horizontal",height=150)
        self.left_box_2_3_l = oasysgui.widgetBox(self.left_box_2_3, "", addSpace=False, orientation="vertical")
        self.left_box_2_3_r = oasysgui.widgetBox(self.left_box_2_3, "", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_emittance_h", "\u03B5x [m.rad]",tooltip="electron_beam_emittance_h",labelWidth=75, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_alpha_h",     "\u03B1x",        tooltip="electron_beam_alpha_h",    labelWidth=75, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_beta_h",      "\u03B2x [m]",    tooltip="electron_beam_beta_h",     labelWidth=75, valueType=float, orientation="horizontal")
        # dispersion effects not considered from Twiss, treated separately
        # oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_eta_h",       "\u03B7x",        tooltip="electron_beam_eta_h",      labelWidth=75, valueType=float, orientation="horizontal")
        # oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_etap_h",      "\u03B7'x",       tooltip="electron_beam_etap_h",     labelWidth=75, valueType=float, orientation="horizontal")


        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_emittance_v", "\u03B5y [m.rad]",tooltip="electron_beam_emittance_v",labelWidth=75, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_alpha_v",     "\u03B1y",        tooltip="electron_beam_alpha_v",    labelWidth=75,valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_beta_v",      "\u03B2y [m]",    tooltip="electron_beam_beta_v",     labelWidth=75, valueType=float, orientation="horizontal")
        # dispersion effects not considered from Twiss, treated separately
        # oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_eta_v",       "\u03B7y",        tooltip="electron_beam_eta_v",      labelWidth=75, valueType=float, orientation="horizontal")
        # oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_etap_v",      "\u03B7'y",       tooltip="electron_beam_etap_v",     labelWidth=75, valueType=float, orientation="horizontal")

        self.set_TypeOfProperties()

        gui.rubber(self.controlArea)


    def set_TypeOfProperties(self):
        self.left_box_2_1.setVisible(self.type_of_properties == 0)
        self.left_box_2_2.setVisible(self.type_of_properties == 1)
        self.left_box_2_3.setVisible(self.type_of_properties == 2)
        self.box_energy_spread.setVisible(self.flag_energy_spread == 1)
        self.set_visibility_energy_spread()

    def set_visibility_energy_spread(self): # to be filled in the upper class
        pass


    def check_data(self):
        congruence.checkStrictlyPositiveNumber(self.electron_energy_in_GeV , "Energy")
        congruence.checkStrictlyPositiveNumber(self.electron_energy_spread, "Energy Spread")
        congruence.checkStrictlyPositiveNumber(self.ring_current, "Ring Current")

        if self.type_of_properties == 0:
            congruence.checkPositiveNumber(self.moment_xx   , "Moment xx")
            congruence.checkPositiveNumber(self.moment_xpxp , "Moment xpxp")
            congruence.checkPositiveNumber(self.moment_yy   , "Moment yy")
            congruence.checkPositiveNumber(self.moment_ypyp , "Moment ypyp")
        elif self.type_of_properties == 1:
            congruence.checkPositiveNumber(self.electron_beam_size_h       , "Horizontal Beam Size")
            congruence.checkPositiveNumber(self.electron_beam_divergence_h , "Horizontal Beam Divergence")
            congruence.checkPositiveNumber(self.electron_beam_size_v       , "Vertical Beam Size")
            congruence.checkPositiveNumber(self.electron_beam_divergence_v , "Vertical Beam Divergence")
        elif self.type_of_properties == 2:
            congruence.checkPositiveNumber(self.electron_beam_emittance_h, "Horizontal Beam Emittance")
            congruence.checkPositiveNumber(self.electron_beam_emittance_v, "Vertical Beam Emittance")
            congruence.checkNumber(self.electron_beam_alpha_h, "Horizontal Beam Alpha")
            congruence.checkNumber(self.electron_beam_alpha_v, "Vertical Beam Alpha")
            congruence.checkNumber(self.electron_beam_beta_h, "Horizontal Beam Beta")
            congruence.checkNumber(self.electron_beam_beta_v, "Vertical Beam Beta")
            # congruence.checkNumber(self.electron_beam_eta_h, "Horizontal Beam Dispersion Eta")
            # congruence.checkNumber(self.electron_beam_eta_v, "Vertical Beam Dispersion Eta")
            # congruence.checkNumber(self.electron_beam_etap_h, "Horizontal Beam Dispersion Eta'")
            # congruence.checkNumber(self.electron_beam_etap_v, "Vertical Beam Dispersion Eta'")

        self.check_magnetic_structure()


    def run_shadow4(self):
        raise Exception("To be defined in the superclass")

    def get_electron_beam(self):
        electron_beam = S4ElectronBeam(energy_in_GeV=self.electron_energy_in_GeV,
                                     energy_spread=self.electron_energy_spread,
                                     current=self.ring_current,
                                     # number_of_bunches=self.number_of_bunches,
                                     )


        recalculate_fields = False

        if self.type_of_properties == 0:
            electron_beam.set_moments_horizontal(self.moment_xx, self.moment_xxp, self.moment_xpxp)
            electron_beam.set_moments_vertical(self.moment_yy, self.moment_yyp, self.moment_ypyp)

            if recalculate_fields:

                x, xp, y, yp = electron_beam.get_sigmas_all()

                self.electron_beam_size_h = x
                self.electron_beam_size_v = y
                self.electron_beam_divergence_h = xp
                self.electron_beam_divergence_v = yp

                twiss_all = electron_beam.get_twiss_no_dispersion_all()
                self.electron_beam_emittance_h = twiss_all[0]
                self.electron_beam_alpha_h     = twiss_all[1]
                self.electron_beam_beta_h      = twiss_all[2]
                # self.electron_beam_eta_h       = 0.0
                # self.electron_beam_etap_h      = 0.0
                self.electron_beam_emittance_v = twiss_all[3]
                self.electron_beam_alpha_v     = twiss_all[4]
                self.electron_beam_beta_v      = twiss_all[5]
                # self.electron_beam_eta_v       = 0.0
                # self.electron_beam_etap_v      = 0.0


        elif self.type_of_properties == 1:
            electron_beam.set_sigmas_all(sigma_x=self.electron_beam_size_h,
                                         sigma_xp=self.electron_beam_divergence_h,
                                         sigma_y=self.electron_beam_size_v,
                                         sigma_yp=self.electron_beam_divergence_v)

            if recalculate_fields:
                moments_all = electron_beam.get_moments_all()

                self.moment_xx   = moments_all[0]
                self.moment_xxp  = moments_all[1]
                self.moment_xpxp = moments_all[2]
                self.moment_yy   = moments_all[3]
                self.moment_yy   = moments_all[4]
                self.moment_ypyp = moments_all[5]

                twiss_all = electron_beam.get_twiss_no_dispersion_all()
                self.electron_beam_emittance_h = twiss_all[0]
                self.electron_beam_alpha_h     = twiss_all[1]
                self.electron_beam_beta_h      = twiss_all[2]
                # self.electron_beam_eta_h       = 0.0
                # self.electron_beam_etap_h      = 0.0
                self.electron_beam_emittance_v = twiss_all[3]
                self.electron_beam_alpha_v     = twiss_all[4]
                self.electron_beam_beta_v      = twiss_all[5]
                # self.electron_beam_eta_v       = 0.0
                # self.electron_beam_etap_v      = 0.0

        elif self.type_of_properties == 2:
            electron_beam.set_twiss_horizontal(self.electron_beam_emittance_h,
                                             self.electron_beam_alpha_h,
                                             self.electron_beam_beta_h,
                                             0.0,
                                             0.0,
                                               )
            electron_beam.set_twiss_vertical(self.electron_beam_emittance_v,
                                             self.electron_beam_alpha_v,
                                             self.electron_beam_beta_v,
                                             0.0,
                                             0.0,
                                             )



            if recalculate_fields:
                x, xp, y, yp = electron_beam.get_sigmas_all()

                self.electron_beam_size_h = x
                self.electron_beam_size_v = y
                self.electron_beam_divergence_h = xp
                self.electron_beam_divergence_v = yp

                moments_all = electron_beam.get_moments_all()

                self.moment_xx   = moments_all[0]
                self.moment_xxp  = moments_all[1]
                self.moment_xpxp = moments_all[2]
                self.moment_yy   = moments_all[3]
                self.moment_yy   = moments_all[4]
                self.moment_ypyp = moments_all[5]
        elif self.type_of_properties == 3:
            electron_beam.set_moments_all(0,0,0,0,0,0)

        return electron_beam

    def populate_fields_from_electron_beam(self, electron_beam):

        self.electron_energy_in_GeV = electron_beam.energy()
        self.electron_energy_spread = electron_beam._energy_spread
        self.ring_current = electron_beam.current()

        moment_xx, moment_xxp, moment_xpxp, moment_yy, moment_yyp, moment_ypyp = electron_beam.get_moments_all()
        self.moment_xx   = moment_xx
        self.moment_xxp  = moment_xxp
        self.moment_xpxp = moment_xpxp
        self.moment_yy   = moment_yy
        self.moment_yyp  = moment_yyp
        self.moment_ypyp = moment_ypyp

        x, xp, y, yp = electron_beam.get_sigmas_all()
        self.electron_beam_size_h       = x
        self.electron_beam_size_v       = y
        self.electron_beam_divergence_h = xp
        self.electron_beam_divergence_v = yp

        twiss_all = electron_beam.get_twiss_no_dispersion_all()
        self.electron_beam_emittance_h = twiss_all[0]
        self.electron_beam_alpha_h = twiss_all[1]
        self.electron_beam_beta_h = twiss_all[2]
        # self.electron_beam_eta_h = 0.0
        # self.electron_beam_etap_h = 0.0
        self.electron_beam_emittance_v = twiss_all[3]
        self.electron_beam_alpha_v = twiss_all[4]
        self.electron_beam_beta_v = twiss_all[5]
        # self.electron_beam_eta_v = 0.0
        # self.electron_beam_etap_v = 0.0

        self.type_of_properties = 1
        self.set_TypeOfProperties()



import numpy
from orangewidget import gui
from orangewidget.settings import Setting

from oasys2.widget import gui as oasysgui
from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from syned.beamline.element_coordinates import ElementCoordinates

from shadow4.beamline.optical_elements.ideal_elements.s4_ideal_fzp import S4IdealFZPElement, S4IdealFZP

from orangecontrib.shadow4.widgets.gui.ow_optical_element import OWOpticalElement

class OWIdealFZP(OWOpticalElement):
    name        = "Ideal FZP"
    description = "Shadow Ideal FZP"
    icon        = "icons/ideal_fzp.png"

    priority = 3.3

    focusing_direction = Setting(3)
    focal = Setting(1.0)
    nominal_wavelength_A = Setting(1.54)  # nominal wavelength in m
    diameter_microns = Setting(500.0)  # FZP diameter

    def __init__(self):
        super().__init__(has_footprint=True)

    def populate_tab_position(self, tab_position):
        self.orientation_box = oasysgui.widgetBox(tab_position, "Optical Element Orientation", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(self.orientation_box, self, "source_plane_distance", "Source Plane Distance [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="source_plane_distance")
        oasysgui.lineEdit(self.orientation_box, self, "image_plane_distance", "Image Plane Distance [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="image_plane_distance")

    def create_basic_settings_subtabs(self, tabs_basic_settings):
        return oasysgui.createTabPage(tabs_basic_settings, "Ideal FZP Parameters")  # to be populated

    def populate_basic_setting_subtabs(self, basic_setting_subtabs):
        tab_ideal_fzp = basic_setting_subtabs

        box_ideal_fzp = oasysgui.widgetBox(tab_ideal_fzp, "", addSpace=False, orientation="vertical", height=180)

        gui.comboBox(box_ideal_fzp, self, "focusing_direction", label="Focusing direction", labelWidth=350,
                     items=["None", "1D: x (sagittal)", "1D: z (meridional)", "2D (radial)"],
                     sendSelectedValue=False,
                     orientation="horizontal",
                     tooltip="focusing_direction")

        gui.separator(box_ideal_fzp)

        oasysgui.lineEdit(box_ideal_fzp, self, "focal", "Design focal distance [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="focal")

        oasysgui.lineEdit(box_ideal_fzp, self, "nominal_wavelength_A", "Design nominal wavelengt [A]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="nominal_wavelength_A")

        oasysgui.lineEdit(box_ideal_fzp, self, "diameter_microns", "FZP diameter [microns]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="diameter_microns")

    def get_coordinates_instance(self):
        return ElementCoordinates(
                p=self.source_plane_distance,
                q=self.image_plane_distance,
                angle_radial=0.0,
                angle_azimuthal=0.0,
                angle_radial_out=numpy.pi,
                )

    def get_optical_element_instance(self):
        try:    name = self.getNode().title
        except: name = "Ideal FZP"

        return S4IdealFZP(name=name,
                focusing_direction=self.focusing_direction,
                nominal_wavelength=self.nominal_wavelength_A * 1e-10,
                focal=self.focal,
                diameter=self.diameter_microns * 1e-6,
                )

    def get_beamline_element_instance(self):
        return S4IdealFZPElement()

add_widget_parameters_to_module(__name__)

'''if __name__ == "__main__":
    from shadow4.beamline.s4_beamline import S4Beamline
    from orangecontrib.shadow4.util.shadow4_objects import ShadowData
    import sys
    def get_test_beam():
        from shadow4.sources.source_geometrical.source_geometrical import SourceGeometrical
        light_source = SourceGeometrical(name='SourceGeometrical', nrays=5000, seed=5676561)
        light_source.set_spatial_type_rectangle(width=0.001000, height=0.001000)
        light_source.set_angular_distribution_flat(hdiv1=0.000000, hdiv2=0.000000, vdiv1=0.000000, vdiv2=0.000000)
        light_source.set_energy_distribution_singleline(5000.000000, unit='A')
        light_source.set_polarization(polarization_degree=1.000000, phase_diff=0.000000, coherent_beam=0)
        beam = light_source.get_beam()

        return ShadowData(beam=beam, beamline=S4Beamline(light_source=light_source))

    from PyQt5.QtWidgets import QApplication
    a = QApplication(sys.argv)
    ow = OWIdealFZP()
    ow.set_shadow_data(get_test_beam())
    # ow.run_shadow4()

    ow.show()
    a.exec_()
    ow.saveSettings()'''



import sys

from PyQt5.QtGui import QPalette, QColor, QFont

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output

from oasys2.widget import gui as oasysgui
from oasys2.widget.widget import OWAction
from oasys2.widget.util.widget_util import EmittingStream
from oasys2.canvas.util.canvas_util import add_widget_parameters_to_module

from orangecontrib.shadow4.widgets.gui.ow_generic_element import GenericElement

from syned.widget.widget_decorator import WidgetDecorator

from shadow4.beamline.optical_elements.ideal_elements.s4_beam_movement import S4BeamMovement, S4BeamMovementElement
from shadow4.tools.logger import set_verbose

from orangecontrib.shadow4.util.shadow4_objects import ShadowData
from orangecontrib.shadow4.util.shadow4_util import ShadowCongruence

class OWBeamMovement(GenericElement, WidgetDecorator):
    name = "Beam movements"
    description = "Shadow Beam Movement"
    icon = "icons/beam_movement.png"
    priority = 10.1

    class Inputs:
        shadow_data = Input("Shadow Data", ShadowData, default=True, auto_summary=False)
        syned_data  = WidgetDecorator.syned_input_data(multi_input=True)

    class Outputs:
        shadow_data = Output("Shadow Data", ShadowData, id="Shadow Data", default=True, auto_summary=False)

    #########################################################
    # Position
    #########################################################
    apply_flag = Setting(1)
    translation_x  = Setting(0.0)
    translation_y  = Setting(0.0)
    translation_z  = Setting(0.0)
    rotation_x     = Setting(0.0)
    rotation_y     = Setting(0.0)
    rotation_z     = Setting(0.0)

    input_data = None
    beamline = None


    def __init__(self):
        super().__init__(show_automatic_box=True, has_footprint=False)

        #
        # main buttons
        #
        self.runaction = OWAction("Run Shadow4/Trace", self)
        self.runaction.triggered.connect(self.run_shadow4)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Run shadow4/trace", callback=self.run_shadow4)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Fields", callback=self.call_reset_settings)
        font = QFont(button.font())
        font.setItalic(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Red'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)
        button.setFixedWidth(150)

        #
        # tabs
        #
        self.tabs_control_area = oasysgui.tabWidget(self.controlArea)
        self.tabs_control_area.setFixedHeight(self.TABS_AREA_HEIGHT)
        self.tabs_control_area.setFixedWidth(self.CONTROL_AREA_WIDTH-5)


        self.tab_movement = oasysgui.createTabPage(self.tabs_control_area, "Movement")           # to be populated

        #
        # populate tabs with widgets
        #

        #########################################################
        # Position
        #########################################################
        self.populate_tab_movement()

        self.update_panels()
        #
        #
        #

        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)

    def resetSettings(self):
        self.translation_x = 0.0
        self.translation_y = 0.0
        self.translation_z = 0.0
        self.rotation_x = 0.0
        self.rotation_y = 0.0
        self.rotation_z = 0.0


    def populate_tab_movement(self):

        gui.comboBox(self.tab_movement, self, "apply_flag", label="Apply movements",
                     labelWidth=250, items=["Off", "On"], callback=self.update_panels,
                     sendSelectedValue=False, orientation="horizontal", tooltip="angles_respect_to")


        self.translation_box = oasysgui.widgetBox(self.tab_movement, "Translation", addSpace=True, orientation="vertical")
        self.rotation_box = oasysgui.widgetBox(self.tab_movement, "Rotation", addSpace=True, orientation="vertical")


        oasysgui.lineEdit(self.translation_box, self, "translation_x", "Translation along X [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="translation_x")
        oasysgui.lineEdit(self.translation_box, self, "translation_y", "Translation along Y [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="translation_y")
        oasysgui.lineEdit(self.translation_box, self, "translation_z", "Translation along Z [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="translation_z")

        oasysgui.lineEdit(self.rotation_box, self, "rotation_x", "1) Rotation along X [rad]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="rotation_x")
        oasysgui.lineEdit(self.rotation_box, self, "rotation_y", "2) Rotation along Y [rad]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="rotation_y")
        oasysgui.lineEdit(self.rotation_box, self, "rotation_z", "3) Rotation along Z [rad]", labelWidth=260,
                          valueType=float, orientation="horizontal", tooltip="rotation_z")


    def update_panels(self):

        if self.apply_flag:
            self.translation_box.setVisible(True)
            self.rotation_box.setVisible(True)
        else:
            self.translation_box.setVisible(False)
            self.rotation_box.setVisible(False)


    def get_oe_instance(self):
        return S4BeamMovement(
            apply_flag = self.apply_flag,
            translation_x = self.translation_x,
            translation_y = self.translation_y,
            translation_z = self.translation_z,
            rotation_x = self.rotation_x,
            rotation_y = self.rotation_y,
            rotation_z = self.rotation_z,
        )

    def get_element_instance(self):
        optical_element = S4BeamMovementElement()
        optical_element.set_optical_element(self.get_oe_instance())
        optical_element.set_input_beam(self.input_data.beam)
        return optical_element

    @Inputs.shadow_data
    def set_shadow_data(self, input_data):
        self.not_interactive = self._check_not_interactive_conditions(input_data)

        self._on_receiving_input()

        if ShadowCongruence.check_empty_data(input_data):
            self.input_data = input_data.duplicate()
            if self.is_automatic_run: self.run_shadow4()

    @Inputs.syned_data
    def set_syned_data(self, index, syned_data):
        self.receive_syned_data(syned_data)

    @Inputs.syned_data.insert
    def insert_syned_data(self, index, syned_data):
        self.receive_syned_data(syned_data)

    @Inputs.syned_data.remove
    def remove_syned_data(self, index):
        pass

    def receive_syned_data(self, data): raise Exception("Not yet implemented")

    def run_shadow4(self):
        if self.input_data is None:
            self.prompt_exception(ValueError("No input beam"))
            return

        try:
            self.progressBarInit()
            set_verbose()
            self.shadow_output.setText("")

            sys.stdout = EmittingStream(textWritten=self._write_stdout)

            element = self.get_element_instance()
            print(element.info())

            output_beam, _ = element.trace_beam()

            beamline = self.input_data.beamline.duplicate()
            beamline.append_beamline_element(element)

            self._set_plot_quality()

            self._plot_results(output_beam, None, progressBarValue=80)

            self.progressBarFinished()

            #
            # script
            #
            script = beamline.to_python_code()
            script += "\n\n\n# test plot"
            script += "\nif True:"
            script += "\n   from srxraylib.plot.gol import plot_scatter"
            script += "\n   # plot_scatter(beam.get_photon_energy_eV(nolost=1), beam.get_column(23, nolost=1), title='(Intensity,Photon Energy)', plot_histograms=0)"
            script += "\n   plot_scatter(1e6 * beam.get_column(1, nolost=1), 1e6 * beam.get_column(3, nolost=1), title='(X,Z) in microns')"
            self.shadow4_script.set_code(script)

            #
            # send beam
            #
            self.Outputs.shadow_data.send(ShadowData(beam=output_beam, beamline=beamline))
        except Exception as exception:
            try:    self._initialize_tabs()
            except: pass
            self.prompt_exception(exception)


add_widget_parameters_to_module(__name__)

'''
if __name__ == "__main__":
    from shadow4.beamline.s4_beamline import S4Beamline
    def get_test_beam():
        # electron beam
        from shadow4.sources.s4_electron_beam import S4ElectronBeam
        electron_beam = S4ElectronBeam(energy_in_GeV=6, energy_spread=0.001, current=0.2)
        electron_beam.set_sigmas_all(sigma_x=3.01836e-05, sigma_y=3.63641e-06, sigma_xp=4.36821e-06,
                                     sigma_yp=1.37498e-06)

        # magnetic structure
        from shadow4.sources.undulator.s4_undulator_gaussian import S4UndulatorGaussian
        source = S4UndulatorGaussian(
            period_length=0.0164,  # syned Undulator parameter (length in m)
            number_of_periods=121.951,  # syned Undulator parameter
            photon_energy=11270,  # Photon energy (in eV)
            delta_e=0.0,  # Photon energy width (in eV)
            ng_e=100,  # Photon energy scan number of points
            flag_emittance=1,  # when sampling rays: Use emittance (0=No, 1=Yes)
            flag_energy_spread=0,  # when sampling rays: Use e- energy spread (0=No, 1=Yes)
            harmonic_number=1,  # harmonic number
            flag_autoset_flux_central_cone=0,  # value to set the flux peak
            flux_central_cone=10000000000.0,  # value to set the flux peak
        )

        # light source
        from shadow4.sources.undulator.s4_undulator_gaussian_light_source import S4UndulatorGaussianLightSource
        light_source = S4UndulatorGaussianLightSource(name='GaussianUndulator', electron_beam=electron_beam,
                                                      magnetic_structure=source, nrays=5000, seed=5676561)
        beam = light_source.get_beam()

        return ShadowData(beam=beam, beamline=S4Beamline(light_source=light_source))

    from PyQt5.QtWidgets import QApplication
    a = QApplication(sys.argv)
    ow = OWBeamMovement()
    ow.set_shadow_data(get_test_beam())
    ow.show()
    a.exec_()
    ow.saveSettings()
'''